import json
import os
from pathlib import Path

import crowsetta
import joblib
import numpy as np
import pandas as pd
from tqdm import tqdm
import torch.utils.data

from .. import (
    constants,
    files,
    io,
    labeled_timebins,
)
from ..logging import log_or_print
from .. import models
from .. import transforms
from ..datasets import VocalDataset
from ..device import get_default as get_default_device


def predict(
    csv_path,
    checkpoint_path,
    labelmap_path,
    model_config_map,
    window_size,
    num_workers=2,
    spect_key="s",
    timebins_key="t",
    spect_scaler_path=None,
    device=None,
    annot_csv_filename=None,
    output_dir=None,
    min_segment_dur=None,
    majority_vote=False,
    save_net_outputs=False,
    logger=None,
):
    """make predictions on dataset with trained model specified in config.toml file.
     Function called by command-line interface.

     Parameters
     ----------
     csv_path : str
         path to where dataset was saved as a csv.
     checkpoint_path : str
         path to directory with checkpoint files saved by Torch, to reload model
     labelmap_path : str
         path to 'labelmap.json' file.
     model_config_map : dict
         where each key-value pair is model name : dict of config parameters
     window_size : int
         size of windows taken from spectrograms, in number of time bins,
         shown to neural networks
     num_workers : int
         Number of processes to use for parallel loading of data.
         Argument to torch.DataLoader. Default is 2.
     spect_key : str
         key for accessing spectrogram in files. Default is 's'.
     timebins_key : str
         key for accessing vector of time bins in files. Default is 't'.
     device : str
         Device on which to work with model + data.
         Defaults to 'cuda' if torch.cuda.is_available is True.
     spect_scaler_path : str
         path to a saved SpectScaler object used to normalize spectrograms.
         If spectrograms were normalized and this is not provided, will give
         incorrect results.
     annot_csv_filename : str
         name of .csv file containing predicted annotations.
         Default is None, in which case the name of the dataset .csv
         is used, with '.annot.csv' appended to it.
     output_dir : str, Path
         path to location where .csv containing predicted annotation
         should be saved. Defaults to current working directory.
     min_segment_dur : float
         minimum duration of segment, in seconds. If specified, then
         any segment with a duration less than min_segment_dur is
         removed from lbl_tb. Default is None, in which case no
         segments are removed.
     majority_vote : bool
         if True, transform segments containing multiple labels
         into segments with a single label by taking a "majority vote",
         i.e. assign all time bins in the segment the most frequently
         occurring label in the segment. This transform can only be
         applied if the labelmap contains an 'unlabeled' label,
         because unlabeled segments makes it possible to identify
         the labeled segments. Default is False.
    save_net_outputs : bool
         if True, save 'raw' outputs of neural networks
         before they are converted to annotations. Default is False.
         Typically the output will be "logits"
         to which a softmax transform might be applied.
         For each item in the dataset--each row in  the `csv_path` .csv--
         the output will be saved in a separate file in `output_dir`,
         with the extension `{MODEL_NAME}.output.npz`. E.g., if the input is a
         spectrogram with `spect_path` filename `gy6or6_032312_081416.npz`,
         and the network is `TweetyNet`, then the net output file
         will be `gy6or6_032312_081416.tweetynet.output.npz`.

     Other Parameters
     ----------------
     logger : logging.Logger
         instance created by vak.logging.get_logger. Default is None.

     Returns
     -------
     None
    """
    if output_dir is None:
        output_dir = Path(os.getcwd())
    else:
        output_dir = Path(output_dir)

    if not output_dir.is_dir():
        raise NotADirectoryError(
            f"value specified for output_dir is not recognized as a directory: {output_dir}"
        )

    if device is None:
        device = get_default_device()

    # ---------------- load data for prediction ------------------------------------------------------------------------
    if spect_scaler_path:
        log_or_print(
            f"loading SpectScaler from path: {spect_scaler_path}",
            logger=logger,
            level="info",
        )
        spect_standardizer = joblib.load(spect_scaler_path)
    else:
        log_or_print(
            f"Not loading SpectScaler, no path was specified",
            logger=logger,
            level="info",
        )
        spect_standardizer = None

    item_transform = transforms.get_defaults(
        "predict",
        spect_standardizer,
        window_size=window_size,
        return_padding_mask=True,
    )

    log_or_print(
        f"loading labelmap from path: {labelmap_path}", logger=logger, level="info"
    )
    with labelmap_path.open("r") as f:
        labelmap = json.load(f)

    log_or_print(
        f"loading dataset to predict from csv path: {csv_path}",
        logger=logger,
        level="info",
    )
    pred_dataset = VocalDataset.from_csv(
        csv_path=csv_path,
        split="predict",
        labelmap=labelmap,
        spect_key=spect_key,
        timebins_key=timebins_key,
        item_transform=item_transform,
    )

    pred_data = torch.utils.data.DataLoader(
        dataset=pred_dataset,
        shuffle=False,
        # batch size 1 because each spectrogram reshaped into a batch of windows
        batch_size=1,
        num_workers=num_workers,
    )

    # ---------------- set up to convert predictions to annotation files -----------------------------------------------
    if annot_csv_filename is None:
        annot_csv_filename = Path(csv_path).stem + constants.ANNOT_CSV_SUFFIX
    annot_csv_path = Path(output_dir).joinpath(annot_csv_filename)
    log_or_print(
        f"will save annotations in .csv file: {annot_csv_path}",
        logger=logger,
        level="info",
    )

    dataset_df = pd.read_csv(csv_path)
    timebin_dur = io.dataframe.validate_and_get_timebin_dur(dataset_df)
    log_or_print(
        f"dataset has timebins with duration: {timebin_dur}",
        logger=logger,
        level="info",
    )

    # ---------------- do the actual predicting + converting to annotations --------------------------------------------
    input_shape = pred_dataset.shape
    # if dataset returns spectrogram reshaped into windows,
    # throw out the window dimension; just want to tell network (channels, height, width) shape
    if len(input_shape) == 4:
        input_shape = input_shape[1:]
    log_or_print(
        f"shape of input to networks used for predictions: {input_shape}",
        logger=logger,
        level="info",
    )

    log_or_print(
        f"instantiating models from model-config map:/n{model_config_map}",
        logger=logger,
        level="info",
    )
    models_map = models.from_model_config_map(
        model_config_map, num_classes=len(labelmap), input_shape=input_shape
    )
    for model_name, model in models_map.items():
        # ---------------- do the actual predicting --------------------------------------------------------------------
        log_or_print(
            f"loading checkpoint for {model_name} from path: {checkpoint_path}",
            logger=logger,
            level="info",
        )
        model.load(checkpoint_path, device=device)
        log_or_print(
            f"running predict method of {model_name}", logger=logger, level="info"
        )
        pred_dict = model.predict(pred_data=pred_data, device=device)

        # ----------------  converting to annotations ------------------------------------------------------------------
        progress_bar = tqdm(pred_data)

        annots = []
        log_or_print(
            "converting predictions to annotations", logger=logger, level="info"
        )
        for ind, batch in enumerate(progress_bar):
            padding_mask, spect_path = batch["padding_mask"], batch["spect_path"]
            padding_mask = np.squeeze(padding_mask)
            if isinstance(spect_path, list) and len(spect_path) == 1:
                spect_path = spect_path[0]
            y_pred = pred_dict[spect_path]

            if save_net_outputs:
                # not sure if there's a better way to get outputs into right shape;
                # can't just call y_pred.reshape() because that basically flattens the whole array first
                # meaning we end up with elements in the wrong order
                # so instead we convert to sequence then stack horizontally, on column axis
                net_output = torch.hstack(y_pred.unbind())
                net_output = net_output[:, padding_mask]
                net_output = net_output.cpu().numpy()
                net_output_path = output_dir.joinpath(
                    Path(spect_path).stem + f"{model_name}{constants.NET_OUTPUT_SUFFIX}"
                )
                np.savez(net_output_path, net_output)

            y_pred = torch.argmax(y_pred, dim=1)  # assumes class dimension is 1
            y_pred = torch.flatten(y_pred).cpu().numpy()[padding_mask]

            spect_dict = files.spect.load(spect_path)
            t = spect_dict[timebins_key]
            labels, onsets_s, offsets_s = labeled_timebins.lbl_tb2segments(
                y_pred,
                labelmap=labelmap,
                t=t,
                min_segment_dur=min_segment_dur,
                majority_vote=majority_vote,
            )
            if labels is None and onsets_s is None and offsets_s is None:
                # handle the case when all time bins are predicted to be unlabeled
                # see https://github.com/NickleDave/vak/issues/383
                continue
            seq = crowsetta.Sequence.from_keyword(
                labels=labels, onsets_s=onsets_s, offsets_s=offsets_s
            )

            audio_fname = files.spect.find_audio_fname(spect_path)
            annot = crowsetta.Annotation(
                seq=seq, audio_path=audio_fname, annot_path=annot_csv_path.name
            )
            annots.append(annot)

        crowsetta.csv.annot2csv(annot=annots, csv_filename=annot_csv_path)



"""fixtures relating to .toml configuration files"""
import json
import shutil

import pytest
import toml


@pytest.fixture
def test_configs_root(test_data_root):
    """Path that points to data_for_tests/configs

    Two types of config files in this directory:
    1) those used by the tests/scripts/generate_data_for_tests.py script.
       Will be listed in configs.json. See ``specific_config`` fixture below
       for details about types of configs.
    2) those used by tests that are static, e.g., ``invalid_section_config.toml``

    This fixture facilitates access to type (2), e.g. in test_config/test_parse
    """
    return test_data_root.joinpath("configs")


@pytest.fixture
def list_of_schematized_configs(test_configs_root):
    """returns list of configuration files,
    schematized with attributes
    so that fixtures and unit tests can specify those attributes
    to find the filename of a specific configuration file,
    that can then be used to get that file.

    Each element in the list is a dict with the following keys:
    `filename`, `config_type`, `audio_format`, `spect_format`, `annot_format`
    These keys define the schema for config files.

    For example, here is the first one:
    {
      "filename": "test_eval_audio_cbin_annot_notmat.toml",
      "config_type": "eval",
      "audio_format": "cbin",
      "spect_format": null,
      "annot_format": "notmat"
    }

    The ``specific_config`` factory fixture returns a function that
    itself return a configuration ``filename``, when provided values for
    all of the other keys.
    """
    with test_configs_root.joinpath("configs.json").open("r") as fp:
        return json.load(fp)["configs"]


@pytest.fixture
def config_that_doesnt_exist(tmp_path):
    return tmp_path / "config_that_doesnt_exist.toml"


@pytest.fixture
def invalid_section_config_path(test_configs_root):
    return test_configs_root.joinpath("invalid_section_config.toml")


@pytest.fixture
def invalid_option_config_path(test_configs_root):
    return test_configs_root.joinpath("invalid_option_config.toml")


@pytest.fixture
def generated_test_configs_root(generated_test_data_root):
    return generated_test_data_root.joinpath("configs")


# ---- path to config files ----
@pytest.fixture
def all_generated_configs(generated_test_configs_root):
    return sorted(generated_test_configs_root.glob("*toml"))


@pytest.fixture
def specific_config(generated_test_configs_root, list_of_schematized_configs, tmp_path):
    """returns a factory function
    that will return the path
    to a specific configuration file, determined by
    characteristics specified by the caller:
    `config_type`, `audio_format`, `spect_format`, `annot_format`

    The factory function actually returns a copy,
    that will be copied into ``tmp_path``,
    so the original remains umodified.

    If ``root_results_dir`` argument is specified
    when calling the factory function,
    it will convert the value for that option in the section
    corresponding to ``config_type`` to the value
    specified for ``root_results_dir``.
    This makes it possible to dynamically change the ``root_results_dir``
    e.g. to the ``tmp_path`` fixture used by unit tests
    """

    def _specific_config(
        config_type,
        model,
        annot_format,
        audio_format=None,
        spect_format=None,
        options_to_change=None,
    ):
        """returns path to a specific configuration file,
        determined by characteristics specified by the caller:
        `config_type`, `audio_format`, `spect_format`, `annot_format`

        Parameters
        ----------
        config_type : str
            corresponding to a `vak` cli command
        annot_format : str
            annotation format, recognized by ``crowsetta``
        audio_format : str
        spect_format : str
        options_to_change : list, dict
            list of dicts with keys 'section', 'option', and 'value'.
            Can be a single dict, in which case only that option is changed.

        Returns
        -------
        config_path : pathlib.Path
            that points to temporary copy of specified config,
            with any options changed as specified
        """
        original_config_path = None

        for schematized_config in list_of_schematized_configs:
            if all(
                [
                    schematized_config["config_type"] == config_type,
                    schematized_config["model"] == model,
                    schematized_config["annot_format"] == annot_format,
                    schematized_config["audio_format"] == audio_format,
                    schematized_config["spect_format"] == spect_format,
                ]
            ):
                original_config_path = generated_test_configs_root.joinpath(
                    schematized_config["filename"]
                )

        if original_config_path is None:
            raise ValueError(
                f"did not find a specific config with `config_type`='{config_type}', "
                f"`annot_format`='{annot_format}', `audio_format`='{audio_format}', "
                f"and `config_type`='{spect_format}', "
            )
        config_copy_path = tmp_path.joinpath(original_config_path.name)
        config_copy_path = shutil.copy(src=original_config_path, dst=config_copy_path)

        if options_to_change is not None:
            if isinstance(options_to_change, dict):
                options_to_change = [options_to_change]
            elif isinstance(options_to_change, list):
                pass
            else:
                raise TypeError(
                    f"invalid type for `options_to_change`: {type(options_to_change)}"
                )

            with config_copy_path.open("r") as fp:
                config_toml = toml.load(fp)

            for opt_dict in options_to_change:
                config_toml[opt_dict["section"]][opt_dict["option"]] = opt_dict["value"]

            with config_copy_path.open("w") as fp:
                toml.dump(config_toml, fp)

        return config_copy_path

    return _specific_config


@pytest.fixture
def all_generated_train_configs(generated_test_configs_root):
    return sorted(generated_test_configs_root.glob("test_train*toml"))


@pytest.fixture
def all_generated_learncurve_configs(generated_test_configs_root):
    return sorted(generated_test_configs_root.glob("test_learncurve*toml"))


@pytest.fixture
def all_generated_eval_configs(generated_test_configs_root):
    return sorted(generated_test_configs_root.glob("test_eval*toml"))


@pytest.fixture
def all_generated_predict_configs(generated_test_configs_root):
    return sorted(generated_test_configs_root.glob("test_predict*toml"))


# ----  config toml from paths ----
def _return_toml(toml_path):
    """return config files loaded into dicts with toml library
    used to test functions that parse config sections, taking these dicts as inputs"""
    with toml_path.open("r") as fp:
        config_toml = toml.load(fp)
    return config_toml


@pytest.fixture
def specific_config_toml(specific_config):
    """returns a function that will return a dict
    containing parsed toml from a
    specific configuration file, determined by
    characteristics specified by the caller:
    `config_type`, `audio_format`, `spect_format`, `annot_format`
    """

    def _specific_config_toml(
        config_type,
        model,
        annot_format,
        audio_format=None,
        spect_format=None,
    ):
        config_path = specific_config(
            config_type, model, annot_format, audio_format, spect_format
        )
        return _return_toml(config_path)

    return _specific_config_toml


@pytest.fixture
def all_generated_configs_toml(all_generated_configs):
    return [_return_toml(config) for config in all_generated_configs]


@pytest.fixture
def all_generated_train_configs_toml(all_generated_train_configs):
    return [_return_toml(config) for config in all_generated_train_configs]


@pytest.fixture
def all_generated_learncurve_configs_toml(all_generated_learncurve_configs):
    return [_return_toml(config) for config in all_generated_learncurve_configs]


@pytest.fixture
def all_generated_eval_configs_toml(all_generated_eval_configs):
    return [_return_toml(config) for config in all_generated_eval_configs]


@pytest.fixture
def all_generated_predict_configs_toml(all_generated_predict_configs):
    return [_return_toml(config) for config in all_generated_predict_configs]


# ---- config toml + path pairs ----
@pytest.fixture
def all_generated_configs_toml_path_pairs(all_generated_configs):
    """zip of tuple pairs: (dict, pathlib.Path)
    where ``Path`` is path to .toml config file and ``dict`` is
    the .toml config from that path
    loaded into a dict with the ``toml`` library
    """
    return zip(
        [_return_toml(config) for config in all_generated_configs],
        all_generated_configs,
    )


@pytest.fixture
def all_generated_train_configs_toml_path_pairs(all_generated_train_configs):
    """zip of tuple pairs: (dict, pathlib.Path)
    where ``Path`` is path to .toml config file and ``dict`` is
    the .toml config from that path
    loaded into a dict with the ``toml`` library
    """
    return zip(
        [_return_toml(config) for config in all_generated_train_configs],
        all_generated_train_configs,
    )


@pytest.fixture
def all_generated_learncurve_configs_toml_path_pairs(all_generated_learncurve_configs):
    """zip of tuple pairs: (dict, pathlib.Path)
    where ``Path`` is path to .toml config file and ``dict`` is
    the .toml config from that path
    loaded into a dict with the ``toml`` library
    """
    return zip(
        [_return_toml(config) for config in all_generated_learncurve_configs],
        all_generated_learncurve_configs,
    )


@pytest.fixture
def all_generated_eval_configs_toml_path_pairs(all_generated_eval_configs):
    """zip of tuple pairs: (dict, pathlib.Path)
    where ``Path`` is path to .toml config file and ``dict`` is
    the .toml config from that path
    loaded into a dict with the ``toml`` library
    """
    return zip(
        [_return_toml(config) for config in all_generated_eval_configs],
        all_generated_eval_configs,
    )


@pytest.fixture
def all_generated_predict_configs_toml_path_pairs(all_generated_predict_configs):
    """zip of tuple pairs: (dict, pathlib.Path)
    where ``Path`` is path to .toml config file and ``dict`` is
    the .toml config from that path
    loaded into a dict with the ``toml`` library
    """
    return zip(
        [_return_toml(config) for config in all_generated_predict_configs],
        all_generated_predict_configs,
    )



import os
from pathlib import Path

import numpy as np
import dask.bag as db
from dask.diagnostics import ProgressBar

from .. import constants
from .. import files
from ..annotation import source_annot_map
from ..converters import labelset_to_set
from ..config.spect_params import SpectParamsConfig
from ..logging import log_or_print
from ..spect import spectrogram


def files_from_dir(audio_dir, audio_format):
    """get all audio files of a given format
    from a directory or its sub-directories,
    using the file extension associated with that annotation format.

    Parameters
    ----------
    audio_dir : str
        path to directory containing audio files.
    audio_format : str
        valid audio file format. One of {'wav', 'cbin'}.

    Returns
    -------
    audio_files : list
        of paths to audio files
    """
    if audio_format not in constants.AUDIO_FORMAT_FUNC_MAP:
        raise ValueError(f"'{audio_format}' is not a valid audio format")
    audio_files = files.from_dir(audio_dir, audio_format)
    return audio_files


def to_spect(
    audio_format,
    spect_params,
    output_dir,
    audio_dir=None,
    audio_files=None,
    annot_list=None,
    audio_annot_map=None,
    labelset=None,
    logger=None,
):
    """makes spectrograms from audio files and saves in array files

    Parameters
    ----------
    audio_format : str
        format of audio files. One of {'wav', 'cbin'}
    spect_params : dict or vak.config.spectrogram.SpectConfig
        parameters for computing spectrogram, from .toml file.
        To see all related parameters, run:
        >>> help(vak.config.spect_params.SpectParamConfig)
        To get a default configuration, create a SpectParamConfig
        with no arguments and then pass that to `to_spect`:
        >>> default_spect_params = vak.config.spect_params.SpectParamConfig()
        >>> to_spect(audio_format='wav', spect_params=default_spect_params, output_dir='.')
    audio_dir : str
        path to directory containing audio files from which to make spectrograms
    audio_files : list
        of str, full paths to audio files from which to make spectrograms
    annot_list : list
        of annotations for array files. Default is None.
    audio_annot_map : dict
        Where keys are paths to array files and value corresponding to each key is
        the annotation for that array file.
        Default is None.
    output_dir : str
        directory in which to save .spect.npz file generated for each audio file.
    labelset : str, list
        of str or int, set of unique labels for vocalizations. Default is None.
        If not None, skip files where the associated annotations contain labels not in ``labelset``.
        ``labelset`` is converted to a Python ``set`` using ``vak.converters.labelset_to_set``.
        See help for that function for details on how to specify labelset.

    Other Parameters
    ----------------
    logger : logging.Logger
        instance created by vak.logging.get_logger. Default is None.

    Returns
    -------
    spect_files : list
        of str, full paths to .spect.npz files

    Notes
    -----
    For each audio file, a corresponding 'spect.npz' file will be created.
    Each '.spect.npz' file contains the following arrays:
        s : numpy.ndarray
            spectrogram, a 2-d array
        f : numpy.ndarray
            vector of centers of frequency bins from spectrogram
        t : numpy.ndarray
            vector of centers of tme bins from spectrogram
        audio_path : numpy.ndarray
            path to source audio file used to create spectrogram

    The names of the arrays are defaults, and will change if different values are specified
    in spect_params for 'spect_key', 'freqbins_key', 'timebins_key', or 'audio_path_key'.
    """
    if audio_format not in constants.VALID_AUDIO_FORMATS:
        raise ValueError(
            f"audio format must be one of '{constants.VALID_AUDIO_FORMATS}'; "
            f"format '{audio_format}' not recognized."
        )

    if all([arg is None for arg in (audio_dir, audio_files, audio_annot_map)]):
        raise ValueError("must specify one of: audio_dir, audio_files, audio_annot_map")

    if audio_dir and audio_files:
        raise ValueError(
            "received values for audio_dir and audio_files, unclear which to use"
        )

    if audio_dir and audio_annot_map:
        raise ValueError(
            "received values for audio_dir and audio_annot_map, unclear which to use"
        )

    if audio_files and audio_annot_map:
        raise ValueError(
            "received values for audio_files and audio_annot_map, unclear which to use"
        )

    if annot_list and audio_annot_map:
        raise ValueError(
            "received values for annot_list and array_annot_map, unclear which annotations to use"
        )

    if labelset is not None:
        labelset = labelset_to_set(labelset)

    if type(spect_params) not in [dict, SpectParamsConfig]:
        raise TypeError(
            "type of spect_params must be an instance of vak.config.spect_params.SpectParamsConfig, "
            "or a dict that can be converted to a SpectParamsConfig instance, "
            f"but was {type(spect_params)}"
        )
    if type(spect_params) is dict:
        spect_params = SpectParamsConfig(**spect_params)

    # validate audio files if supplied by user
    if audio_files:
        # make sure audio files are all the same type, and the same as audio format specified
        exts = []
        for audio_file in audio_files:
            root, ext = os.path.splitext(audio_file)
            ext = (
                ext.lower()
            )  # make case-insensitive, e.g. treat .wav and .WAV the same
            exts.append(ext)
        uniq_ext = set(exts)
        if len(uniq_ext) > 1:
            raise ValueError(
                "audio_files should all have the same extension, "
                f"but found more than one: {uniq_ext}"
            )
        else:
            ext_str = uniq_ext.pop()
            if audio_format not in ext_str:
                raise ValueError(
                    f"audio format. '{audio_format}', does not match extensions in audio_files, '{ext_str}''"
                )

        if annot_list:  # make map we can validate below
            audio_annot_map = source_annot_map(audio_files, annot_list)

    # otherwise get audio files using audio dir (won't need to validate audio files)
    if audio_dir:
        audio_files = files_from_dir(audio_dir, audio_format)
        if annot_list:
            audio_annot_map = source_annot_map(audio_files, annot_list)

    log_or_print("creating array files with spectrograms", logger=logger, level="info")

    # use mapping (if generated/supplied) with labelset, if supplied, to filter
    if audio_annot_map:
        if labelset:  # then remove annotations with labels not in labelset
            # note we do this here so it happens regardless of whether
            # user supplied audio_annot_map *or* we constructed it above
            for audio_file, annot in list(audio_annot_map.items()):
                # loop in a verbose way (i.e. not a comprehension)
                # so we can give user warning when we skip files
                annot_labelset = set(annot.seq.labels)
                # below, set(labels_mapping) is a set of that dict's keys
                if not annot_labelset.issubset(set(labelset)):
                    # because there's some label in labels that's not in labelset
                    audio_annot_map.pop(audio_file)
                    extra_labels = annot_labelset - labelset
                    log_or_print(
                        f"Found labels, {extra_labels}, in {Path(audio_file).name}, "
                        "that are not in labels_mapping. Skipping file.",
                        logger=logger,
                        level="info",
                    )
        audio_files = sorted(list(audio_annot_map.keys()))

    # this is defined here so all other arguments to 'to_spect' are in scope
    def _spect_file(audio_file):
        """helper function that enables parallelized creation of array
        files containing spectrograms.
        Accepts path to audio file, saves .npz file with spectrogram"""
        dat, fs = constants.AUDIO_FORMAT_FUNC_MAP[audio_format](audio_file)
        s, f, t = spectrogram(
            dat,
            fs,
            spect_params.fft_size,
            spect_params.step_size,
            spect_params.thresh,
            spect_params.transform_type,
            spect_params.freq_cutoffs,
        )
        spect_dict = {
            spect_params.spect_key: s,
            spect_params.freqbins_key: f,
            spect_params.timebins_key: t,
            spect_params.audio_path_key: audio_file,
        }
        basename = os.path.basename(audio_file)
        npz_fname = os.path.join(os.path.normpath(output_dir), basename + ".spect.npz")
        np.savez(npz_fname, **spect_dict)
        return npz_fname

    bag = db.from_sequence(audio_files)
    with ProgressBar():
        spect_files = list(bag.map(_spect_file))
    # sort because ordering from Dask not guaranteed
    spect_files = sorted(spect_files)
    return spect_files



"""default item transforms used with the different command-line interface commands

"item" transforms because they apply transforms to input parameters
and then return them in an "item" (dictionary)
that is turn returned by the __getitem__ method of a vak.VocalDataset.
Having the transform return a dictionary makes it possible to avoid
coupling the VocalDataset __getitem__ implementation to the transforms
needed for specific neural network models, e.g., whether the returned
output includes a mask to crop off padding that was added.
"""
import torchvision.transforms

from . import transforms as vak_transforms


class TrainItemTransform:
    """default transform used when training models"""

    def __init__(
        self,
        spect_standardizer=None,
    ):
        if spect_standardizer is not None:
            if isinstance(spect_standardizer, vak_transforms.StandardizeSpect):
                source_transform = [spect_standardizer]
            else:
                raise TypeError(
                    f"invalid type for spect_standardizer: {type(spect_standardizer)}. "
                    "Should be an instance of vak.transforms.StandardizeSpect"
                )
        else:
            source_transform = []

        source_transform.extend(
            [
                vak_transforms.ToFloatTensor(),
                vak_transforms.AddChannel(),
            ]
        )
        self.source_transform = torchvision.transforms.Compose(source_transform)
        self.annot_transform = vak_transforms.ToLongTensor()

    def __call__(self, source, annot, spect_path=None):
        source = self.source_transform(source)
        annot = self.annot_transform(annot)
        item = {
            "source": source,
            "annot": annot,
        }

        if spect_path is not None:
            item["spect_path"] = spect_path

        return item


class EvalItemTransform:
    """default transform used when evaluating models

    Returned item includes "source" spectrogram reshaped into a stack of windows,
    with padded added to make reshaping possible, and annotation also padded and
    reshaped.
    If return_padding_mask is True, item includes 'padding_mask' that
    can be used to crop off any predictions made on the padding.
    """

    def __init__(
        self,
        window_size,
        spect_standardizer=None,
        padval=0.0,
        return_padding_mask=True,
        channel_dim=1,
    ):
        if spect_standardizer is not None:
            if not isinstance(spect_standardizer, vak_transforms.StandardizeSpect):
                raise TypeError(
                    f"invalid type for spect_standardizer: {type(spect_standardizer)}. "
                    "Should be an instance of vak.transforms.StandardizeSpect"
                )
        self.spect_standardizer = spect_standardizer

        self.pad_to_window = vak_transforms.PadToWindow(
            window_size, padval, return_padding_mask=return_padding_mask
        )

        self.source_transform_after_pad = torchvision.transforms.Compose(
            [
                vak_transforms.ViewAsWindowBatch(window_size),
                vak_transforms.ToFloatTensor(),
                # below, add channel at first dimension because windows become batch
                vak_transforms.AddChannel(channel_dim=channel_dim),
            ]
        )

        self.annot_transform = vak_transforms.ToLongTensor()

    def __call__(self, source, annot, spect_path=None):
        if self.spect_standardizer:
            source = self.spect_standardizer(source)

        if self.pad_to_window.return_padding_mask:
            source, padding_mask = self.pad_to_window(source)
        else:
            source = self.pad_to_window(source)
            padding_mask = None
        source = self.source_transform_after_pad(source)

        annot = self.annot_transform(annot)

        item = {
            "source": source,
            "annot": annot,
        }

        if padding_mask is not None:
            item["padding_mask"] = padding_mask

        if spect_path is not None:
            item["spect_path"] = spect_path

        return item


class PredictItemTransform:
    """default transform used when using trained models to make predictions.

    Returned item includes "source" spectrogram reshaped into a stack of windows,
    with padded added to make reshaping possible.
    If return_padding_mask is True, item includes 'padding_mask' that
    can be used to crop off any predictions made on the padding.
    """

    def __init__(
        self,
        window_size,
        spect_standardizer=None,
        padval=0.0,
        return_padding_mask=True,
        channel_dim=1,
    ):
        if spect_standardizer is not None:
            if not isinstance(spect_standardizer, vak_transforms.StandardizeSpect):
                raise TypeError(
                    f"invalid type for spect_standardizer: {type(spect_standardizer)}. "
                    "Should be an instance of vak.transforms.StandardizeSpect"
                )
        self.spect_standardizer = spect_standardizer

        self.pad_to_window = vak_transforms.PadToWindow(
            window_size, padval, return_padding_mask=return_padding_mask
        )

        self.source_transform_after_pad = torchvision.transforms.Compose(
            [
                vak_transforms.ViewAsWindowBatch(window_size),
                vak_transforms.ToFloatTensor(),
                # below, add channel at first dimension because windows become batch
                vak_transforms.AddChannel(channel_dim=channel_dim),
            ]
        )

    def __call__(self, source, spect_path=None):
        if self.spect_standardizer:
            source = self.spect_standardizer(source)

        if self.pad_to_window.return_padding_mask:
            source, padding_mask = self.pad_to_window(source)
        else:
            source = self.pad_to_window(source)
            padding_mask = None

        source = self.source_transform_after_pad(source)

        item = {
            "source": source,
        }

        if padding_mask is not None:
            item["padding_mask"] = padding_mask

        if spect_path is not None:
            item["spect_path"] = spect_path

        return item


def get_defaults(
    mode,
    spect_standardizer=None,
    window_size=None,
    padval=0.0,
    return_padding_mask=False,
):
    """get default transforms

    Parameters
    ----------
    mode : str
        one of {'train', 'eval', 'predict'}. Determines set of transforms.
    spect_standardizer : vak.transforms.StandardizeSpect
        instance that has already been fit to dataset, using fit_df method.
        Default is None, in which case no standardization transform is applied.
    window_size : int
        width of window in number of elements. Argument to PadToWindow transform.
    padval : float
        value to pad with. Added to end of array, the "right side" if 2-dimensional.
        Argument to PadToWindow transform. Default is 0.
    return_padding_mask : bool
        if True, the dictionary returned by ItemTransform classes will include
        a boolean vector to use for cropping back down to size before padding.
        padding_mask has size equal to width of padded array, i.e. original size
        plus padding at the end, and has values of 1 where
        columns in padded are from the original array,
        and values of 0 where columns were added for padding.

    Returns
    -------
    transform, target_transform : callable
        one or more vak transforms to be applied to inputs x and, during training, the target y.
        If more than one transform, they are combined into an instance of torchvision.transforms.Compose.
        Note that when mode is 'predict', the target transform is None.
    """
    # regardless of mode, transform always starts with StandardizeSpect, if used
    if spect_standardizer is not None:
        if not isinstance(spect_standardizer, vak_transforms.StandardizeSpect):
            raise TypeError(
                f"invalid type for spect_standardizer: {type(spect_standardizer)}. "
                "Should be an instance of vak.transforms.StandardizeSpect"
            )

    if mode == "train":
        if spect_standardizer is not None:
            transform = [spect_standardizer]
        else:
            transform = []

        transform.extend(
            [
                vak_transforms.ToFloatTensor(),
                vak_transforms.AddChannel(),
            ]
        )
        transform = torchvision.transforms.Compose(transform)

        target_transform = vak_transforms.ToLongTensor()
        return transform, target_transform

    elif mode == "predict":
        item_transform = PredictItemTransform(
            spect_standardizer=spect_standardizer,
            window_size=window_size,
            padval=padval,
            return_padding_mask=return_padding_mask,
        )
        return item_transform

    elif mode == "eval":
        item_transform = EvalItemTransform(
            spect_standardizer=spect_standardizer,
            window_size=window_size,
            padval=padval,
            return_padding_mask=return_padding_mask,
        )
        return item_transform

    else:
        raise ValueError(f"invalid mode: {mode}")



"""tests for vak.files.spect module"""
from pathlib import Path

import pytest

import vak.files
from vak.constants import VALID_AUDIO_FORMATS


@pytest.mark.parametrize(
    "spect_format",
    [
        "mat",
        "npz",
    ],
)
def test_find_audio_fname_with_mat(spect_format, specific_spect_list):
    """test ```vak.files.spect.find_audio_fname`` works when we give it a list of """
    spect_list = specific_spect_list(spect_format)
    audio_fnames = [
        vak.files.spect.find_audio_fname(spect_path) for spect_path in spect_list
    ]
    for spect_path, audio_fname in zip(spect_list, audio_fnames):
        # make sure we gout out a filename that was actually in spect_path
        assert spect_path.name.startswith(audio_fname)
        # make sure it's some valid audio format
        assert Path(audio_fname).suffix.replace(".", "") in VALID_AUDIO_FORMATS



import pandas as pd

from .. import annotation
from .. import files
from .. import labeled_timebins


class VocalDataset:
    """Base class to represent a dataset of vocalizations.
    It assumes that the dataset consists of spectrograms of
    vocalizations and, optionally, annotations
    for those vocalizations.
    """

    def __init__(
        self,
        csv_path,
        spect_paths,
        annots=None,
        labelmap=None,
        spect_key="s",
        timebins_key="t",
        item_transform=None,
    ):
        """initialize a VocalDataset instance

        Parameters
        ----------
        csv_path : str, Path
            path to a .csv file that represents the dataset.
        spect_paths : numpy.ndarray
            column from DataFrame that represents dataset,
            consisting of paths to files containing spectrograms as arrays
        annots : list
            of crowsetta.Annotation instances,
            loaded from from DataFrame that represents dataset, using vak.annotation.from_df.
            Default is None, in which case no annotation is returned with each item
            in the dataset.
        labelmap : dict
            that maps labels from dataset to a series of consecutive integer.
            To create a label map, pass a set of labels to the `vak.utils.labels.to_map` function.
        spect_key : str
            key to access spectograms in array files. Default is 's'.
        timebins_key : str
            key to access time bin vector in array files. Default is 't'.
        item_transform : callable
            A function / transform that takes an input numpy array or torch Tensor,
            and optionally a target array or Tensor, and returns a dictionary.
            This dictionary is the item returned when indexing into the dataset.
            Default is None.
        """
        self.csv_path = csv_path
        self.spect_paths = spect_paths
        self.spect_key = spect_key
        self.timebins_key = timebins_key
        self.annots = annots
        self.labelmap = labelmap
        if "unlabeled" in self.labelmap:
            self.unlabeled_label = self.labelmap["unlabeled"]
        else:
            # if there is no "unlabeled label" (e.g., because all segments have labels)
            # just assign dummy value that will end up getting replaced by actual labels by label_timebins()
            self.unlabeled_label = 0
        self.item_transform = item_transform

        tmp_x_ind = 0
        tmp_item = self.__getitem__(tmp_x_ind)
        # used by vak functions that need to determine size of input,
        # e.g. when initializing a neural network model
        self.shape = tmp_item["source"].shape

    def __getitem__(self, idx):
        spect_path = self.spect_paths[idx]
        spect_dict = files.spect.load(spect_path)
        spect = spect_dict[self.spect_key]

        if self.annots is not None:
            timebins = spect_dict[self.timebins_key]

            annot = self.annots[idx]
            lbls_int = [self.labelmap[lbl] for lbl in annot.seq.labels]
            # "lbl_tb": labeled timebins. Target for output of network
            lbl_tb = labeled_timebins.label_timebins(
                lbls_int,
                annot.seq.onsets_s,
                annot.seq.offsets_s,
                timebins,
                unlabeled_label=self.unlabeled_label,
            )
            item = self.item_transform(spect, lbl_tb, spect_path)
        else:
            item = self.item_transform(spect, spect_path)

        return item

    def __len__(self):
        """number of batches"""
        return len(self.spect_paths)

    @classmethod
    def from_csv(
        cls,
        csv_path,
        split,
        labelmap,
        spect_key="s",
        timebins_key="t",
        item_transform=None,
    ):
        """given a path to a csv representing a dataset,
        returns an initialized VocalDataset.

        Parameters
        ----------
        csv_path : str, Path
            path to a .csv file that represents the dataset.
        spect_paths : numpy.ndarray
            column from DataFrame that represents dataset,
            consisting of paths to files containing spectrograms as arrays
        annots : list
            of crowsetta.Annotation instances,
            loaded from from DataFrame that represents dataset, using vak.annotation.from_df.
            Default is None, in which case no annotation is returned with each item
            in the dataset.
        labelmap : dict
            that maps labels from dataset to a series of consecutive integer.
            To create a label map, pass a set of labels to the `vak.utils.labels.to_map` function.
        spect_key : str
            key to access spectograms in array files. Default is 's'.
        timebins_key : str
            key to access time bin vector in array files. Default is 't'.
        item_transform : callable
            A function / transform that takes an input numpy array or torch Tensor,
            and optionally a target array or Tensor, and returns a dictionary.
            This dictionary is the item returned when indexing into the dataset.
            Default is None.

        Returns
        -------
        initialized instance of VocalDataset
        """
        df = pd.read_csv(csv_path)
        if not df["split"].str.contains(split).any():
            raise ValueError(f"split {split} not found in dataset in csv: {csv_path}")
        else:
            df = df[df["split"] == split]

        spect_paths = df["spect_path"].values

        # below, annots will be None if no format is specified in the `annot_format` column of the dataframe.
        # this is intended behavior; makes it possible to use same dataset class for prediction
        annots = annotation.from_df(df)

        return cls(
            csv_path,
            spect_paths,
            annots,
            labelmap,
            spect_key,
            timebins_key,
            item_transform,
        )



"""fixtures relating to array files containing spectrograms"""
import pytest

import vak.files.spect


@pytest.fixture
def spect_dir_mat(source_test_data_root):
    return source_test_data_root.joinpath("spect_mat_annot_yarden", "llb3", "spect")


@pytest.fixture
def spect_dir_npz(generated_test_data_root):
    return sorted(
        generated_test_data_root.joinpath(
            "prep", "train", "audio_cbin_annot_notmat"
        ).glob("spectrograms_generated*")
    )[0]


@pytest.fixture
def specific_spect_dir(spect_dir_mat, spect_dir_npz):
    def _specific_spect_dir(spect_format):
        if spect_format == "mat":
            return spect_dir_mat
        elif spect_format == "npz":
            return spect_dir_npz
        else:
            raise ValueError(f"invalid spect_format: {spect_format}")

    return _specific_spect_dir


@pytest.fixture
def spect_list_mat(spect_dir_mat):
    return sorted(spect_dir_mat.glob("*.mat"))


@pytest.fixture
def spect_list_npz(spect_dir_npz):
    return sorted(spect_dir_npz.glob("*.spect.npz"))


@pytest.fixture
def spect_list_mat_all_labels_in_labelset(
    spect_list_mat, annot_list_yarden, labelset_yarden
):
    """list of .mat spectrogram files where all labels in associated annotation **are** in labelset"""
    labelset_yarden = set(labelset_yarden)
    spect_list_labels_in_labelset = []
    for spect_path in spect_list_mat:
        audio_fname = vak.files.spect.find_audio_fname(spect_path)
        annot = [
            annot for annot in annot_list_yarden if annot.audio_path.name == audio_fname
        ]
        assert len(annot) == 1
        annot = annot[0]
        if set(annot.seq.labels).issubset(labelset_yarden):
            spect_list_labels_in_labelset.append(spect_path)

    return spect_list_labels_in_labelset


@pytest.fixture
def spect_list_npz_all_labels_in_labelset(
    spect_list_npz, annot_list_notmat, labelset_notmat
):
    """list of .npz spectrogram files where all labels in associated annotation **are** in labelset"""
    labelset_notmat = set(labelset_notmat)
    spect_list_labels_in_labelset = []
    for spect_path in spect_list_npz:
        audio_fname = vak.files.spect.find_audio_fname(spect_path)
        annot = [
            annot for annot in annot_list_notmat if annot.audio_path.name == audio_fname
        ]
        assert len(annot) == 1
        annot = annot[0]
        if set(annot.seq.labels).issubset(labelset_notmat):
            spect_list_labels_in_labelset.append(spect_path)

    return spect_list_labels_in_labelset


@pytest.fixture
def spect_list_mat_labels_not_in_labelset(
    spect_list_mat, annot_list_yarden, labelset_yarden
):
    """list of .mat spectrogram files where some labels in associated annotation are **not** in labelset"""
    labelset_yarden = set(labelset_yarden)
    spect_list_labels_not_in_labelset = []
    for spect_path in spect_list_mat:
        audio_fname = vak.files.spect.find_audio_fname(spect_path)
        annot = [
            annot for annot in annot_list_yarden if annot.audio_path.name == audio_fname
        ]
        assert len(annot) == 1
        annot = annot[0]
        # notice if labels **not** a subset of labelset
        if not set(annot.seq.labels).issubset(labelset_yarden):
            spect_list_labels_not_in_labelset.append(spect_path)

    err = "not finding .mat spectrogram files where labels in associated annotations are not in dataset"
    assert len(spect_list_labels_not_in_labelset) > 0, err
    return spect_list_labels_not_in_labelset


@pytest.fixture
def spect_list_npz_labels_not_in_labelset(
    spect_list_npz, annot_list_notmat, labelset_notmat
):
    """list of .npz spectrogram files where some labels in associated annotation are  **not** in labelset"""
    labelset_notmat = set(labelset_notmat)
    spect_list_labels_not_in_labelset = []
    for spect_path in spect_list_npz:
        audio_fname = vak.files.spect.find_audio_fname(spect_path)
        annot = [
            annot for annot in annot_list_notmat if annot.audio_path.name == audio_fname
        ]
        assert len(annot) == 1
        annot = annot[0]
        if set(annot.seq.labels).issubset(labelset_notmat):
            spect_list_labels_not_in_labelset.append(spect_path)

    err = "not finding .npz spectrogram files where labels in associated annotations are not in dataset"
    assert len(spect_list_labels_not_in_labelset) > 0, err
    return spect_list_labels_not_in_labelset


@pytest.fixture
def specific_spect_list(
    spect_list_mat,
    spect_list_mat_all_labels_in_labelset,
    spect_list_mat_labels_not_in_labelset,
    spect_list_npz,
    spect_list_npz_all_labels_in_labelset,
    spect_list_npz_labels_not_in_labelset,
):
    def _specific_spect_list(spect_format, qualifier=None):
        MAP = {
            "mat": {
                None: spect_list_mat,
                "all_labels_in_labelset": spect_list_mat_all_labels_in_labelset,
                "labels_not_in_labelset": spect_list_mat_labels_not_in_labelset,
            },
            "npz": {
                None: spect_list_npz,
                "all_labels_in_labelset": spect_list_npz_all_labels_in_labelset,
                "labels_not_in_labelset": spect_list_npz_labels_not_in_labelset,
            },
        }
        return MAP[spect_format][qualifier]

    return _specific_spect_list



from . import functional as F


__all__ = [
    "Accuracy",
]


class Accuracy:
    """standard supervised learning classification accuracy.


    Parameters
    ----------
    y_pred : torch.Tensor
    y_true : torch.Tensor

    Returns
    -------
    acc : float
        between 0 and 1. Sum of predicted labels that equal true labels,
        divided by number of true labels.
    """

    def __init__(self):
        pass

    def __call__(self, y_pred, y_true):
        return F.accuracy(y_pred, y_true)



"""fixtures used to test the vak.split sub-package"""
from evfuncs import load_cbin
from scipy.io import loadmat
import pytest

from vak.timebins import timebin_dur_from_vec


@pytest.fixture
def audio_cbin_annot_notmat_durs_labels(
    audio_list_cbin, annot_list_notmat, labelset_notmat
):
    """returns list of durations of .cbin audio files, and
    another list with labels from annotations in each .not.mat file
    for the associated .cbin audio file"""
    durs = []
    labels = []
    for audio_file, annot in zip(audio_list_cbin, annot_list_notmat):
        if set(annot.seq.labels).issubset(labelset_notmat):
            labels.append(annot.seq.labels)
            data, fs = load_cbin(audio_file)
            durs.append(data.shape[0] / fs)
    return durs, labels


@pytest.fixture
def spect_mat_annot_yarden_durs_labels(
    spect_list_mat, annot_list_yarden, labelset_yarden
):
    """returns list of durations of .mat spectrogram files, and
    another list with labels from annotations in the .mat file
    for each associated .mat spectrogram file"""

    durs = []
    labels = []
    for spect_file_mat, annot in zip(spect_list_mat, annot_list_yarden):
        if set(annot.seq.labels).issubset(labelset_yarden):
            labels.append(annot.seq.labels)
            mat_dict = loadmat(spect_file_mat)
            timebin_dur = timebin_dur_from_vec(mat_dict["t"])
            dur = mat_dict["s"].shape[-1] * timebin_dur
            durs.append(dur)
    return durs, labels



from .__about__ import (
    __author__,
    __commit__,
    __copyright__,
    __email__,
    __license__,
    __summary__,
    __title__,
    __uri__,
    __version__,
)

from . import (
    __main__,
    annotation,
    cli,
    config,
    csv,
    datasets,
    device,
    engine,
    entry_points,
    files,
    io,
    labels,
    labeled_timebins,
    logging,
    metrics,
    models,
    nn,
    plot,
    spect,
    tensorboard,
    timebins,
    transforms,
    split,
    validators,
)

from .engine.model import Model


__all__ = [
    "__main__",
    "annotation",
    "cli",
    "config",
    "csv",
    "datasets",
    "device",
    "engine",
    "entry_points",
    "files",
    "io",
    "labels",
    "labeled_timebins",
    "logging",
    "metrics",
    "Model",
    "models",
    "nn",
    "plot",
    "spect",
    "split",
    "tensorboard",
    "timebins",
    "transforms",
    "validators",
]



import random
import itertools
import logging


def inc_freq(durs, labels, labelset, train_dur, test_dur, val_dur):
    """randomly partition vocalizations into training, test, and validation
    sets

    This algorithm finds the unique set of labels for the vocalizations, then
    sorts them in order of increasing frequency. It then partitions the vocalizations
    into sets starting with the label that occurs *least* frequently. This is a
    'fail-early' approach -- if there's not enough instances of some vocalization,
    we will see that explicitly and be able to error out with an informative message"""
    logger = logging.getLogger(__name__)
    logger.setLevel("INFO")

    # map each label in labelset to list of indices where it occurs in labels
    # (the list of labels from vocalizations)
    labelset_where = {}
    for lbl in labelset:
        labelset_where[lbl] = []
    for lbls_ind, lbls_list in enumerate(labels):
        lbls_set = set(lbls_list)
        for lbl in lbls_set:
            labelset_where[lbl].append(lbls_ind)

    # map this to (duration, labels) tuples for when we do random partitioning in main loop
    labelset_durs_labels = {}
    for lbl, inds in labelset_where.items():
        lbl_durs = [durs[ind] for ind in inds]
        lbl_lbls = [labels[ind] for ind in inds]
        labelset_durs_labels[lbl] = (lbl_durs, lbl_lbls)

    # now count number of occurrences of each label
    labelset_num_occurs = {}
    for lbl, inds in labelset_where.items():
        labelset_num_occurs[lbl] = len(inds)

    labelset_num_occurs = sorted(
        # use value as key for sorting, convert to float so its numeric not alphabetic sort
        labelset_num_occurs.items(),
        key=lambda kv: float(kv[1]),
    )
    # we use this to loop through labelset in order of increasing occurrence below
    labelset_incfreq = [tup[0] for tup in labelset_num_occurs]

    if val_dur is None:
        total_target_dur = sum([train_dur, test_dur])
    elif val_dur is not None:
        total_target_dur = sum([train_dur, test_dur, val_dur])

    iter = 1
    all_labels_err = (
        "Did not successfully divide data into training, "
        "validation, and test sets of sufficient duration "
        "after 1000 iterations."
        " Try increasing the total size of the data set."
    )

    while 1:
        train_inds = []
        test_inds = []
        if val_dur:
            val_inds = []
        else:
            val_inds = None

        total_train_dur = 0
        total_val_dur = 0
        total_test_dur = 0

        choice = ["train", "test"]
        if val_dur:
            choice.append("val")

        least_freq = labelset_num_occurs[0][0]
        least_freq_num = labelset_num_occurs[0][1]
        if least_freq_num < len(choice):
            raise ValueError(
                f"number of occurrences of least frequent label class, {least_freq}, is {least_freq_num}, "
                f"which is less than the number of datasets: {len(choice)}.\nCan not split datasets and have "
                f"each split contain this class."
            )

        while 1:
            for label in labelset_incfreq:
                durs_labels_tup = labelset_durs_labels[label]
                durs_labels_inds = list(range(len(durs_labels_tup[0])))
                random.shuffle(durs_labels_inds)

                # pop durations off list and append to randomly-chosen
                # list, either train, val, or test set.
                # Do this until the total duration for each data set is equal
                # to or greater than the target duration for each set.
                try:
                    ind = durs_labels_inds.pop()
                except IndexError:
                    if len(durs_labels_inds) == 0:
                        logger.debug(
                            "Ran out of elements while dividing dataset into subsets of "
                            f"specified durations. Iteration {iter}"
                        )
                        iter += 1
                        break  # do next iteration
                    else:
                        raise

                which_set = random.randint(0, len(choice) - 1)
                which_set = choice[which_set]
                if which_set == "train":
                    train_inds.append(ind)
                    total_train_dur += durs_labels_list[ind][0]  # ind 0 is duration
                    if total_train_dur >= train_dur:
                        choice.pop(choice.index("train"))
                elif which_set == "val":
                    val_inds.append(ind)
                    total_val_dur += durs_labels_list[ind][0]  # ind 0 is duration
                    if total_val_dur >= val_dur:
                        choice.pop(choice.index("val"))
                elif which_set == "test":
                    test_inds.append(ind)
                    total_test_dur += durs_labels_list[ind][0]  # ind 0 is duration
                    if total_test_dur >= test_dur:
                        choice.pop(choice.index("test"))

                if len(choice) < 1:
                    total_all_durs = total_train_dur + total_test_dur
                    if val_dur:
                        total_all_durs += total_val_dur
                    if total_all_durs < total_target_dur:
                        raise ValueError(
                            f"Loop to find subsets completed but total duration of subsets, {total_all_durs} seconds, "
                            f"is less than total duration specified: {total_target_dur} seconds."
                        )
                    else:
                        break

                if iter > 1000:
                    raise ValueError(
                        "Could not find subsets of sufficient duration in "
                        "less than 1000 iterations."
                    )

        # make sure that each set contains all classes we
        # want the network to learn
        train_tups = [durs_labels_list[ind] for ind in train_inds]  # tup = a tuple
        test_tups = [durs_labels_list[ind] for ind in test_inds]
        if val_dur:
            val_tups = [durs_labels_list[ind] for ind in val_inds]

        train_labels = itertools.chain.from_iterable([tup[1] for tup in train_tups])
        train_labels = set(train_labels)  # make set to get unique values

        test_labels = itertools.chain.from_iterable([tup[1] for tup in test_tups])
        test_labels = set(test_labels)

        if val_dur:
            val_labels = itertools.chain.from_iterable([tup[1] for tup in val_tups])
            val_labels = set(val_labels)

        if train_labels != set(labelset):
            iter += 1
            if iter > 1000:
                raise ValueError(all_labels_err)
            else:
                logger.debug(
                    "Train labels did not contain all labels in labelset. "
                    f"Getting new training set. Iteration {iter}"
                )
                continue
        elif test_labels != set(labelset):
            iter += 1
            if iter > 1000:
                raise ValueError(all_labels_err)
            else:
                logger.debug(
                    "Test labels did not contain all labels in labelset. "
                    f"Getting new test set. Iteration {iter}"
                )
                continue
        elif val_dur is not None and val_labels != set(labelset):
            iter += 1
            if iter > 1000:
                raise ValueError(all_labels_err)
            else:
                logger.debug(
                    "Validation labels did not contain all labels in labelset. "
                    f"Getting new validation set. Iteration {iter}"
                )
                continue

        else:
            break

    return train_inds, test_inds, val_inds



"""tests for vak.config.predict module"""
import vak.config.predict
import vak.split


def test_predict_attrs_class(all_generated_predict_configs_toml):
    """test that instantiating PredictConfig class works as expected"""
    for config_toml in all_generated_predict_configs_toml:
        predict_section = config_toml["PREDICT"]
        config = vak.config.predict.PredictConfig(**predict_section)
        assert isinstance(config, vak.config.predict.PredictConfig)



import torch


def accuracy(y_pred, y_true):
    """standard supervised learning classification accuracy:
    Sum of predicted labels that equal true labels, divided by number of true labels.

    Parameters
    ----------
    y_pred : torch.Tensor
    y_true : torch.Tensor

    Returns
    -------
    acc : float
        between 0 and 1. Sum of predicted labels that equal true labels,
        divided by number of true labels.
    """
    correct = torch.eq(y_pred, y_true).view(-1)
    return correct.sum().item() / correct.shape[0]



"""This script gets run by continuous integration 
(in ./github/workflows/ci-{os}.yml files)
so that all the paths are correct on the virtual machines
"""
from pathlib import Path

import pandas as pd

HERE = Path(__file__).parent
PROJ_ROOT = HERE / ".." / ".."
PROJ_ROOT_ABS = PROJ_ROOT.resolve()  # <- used to fix paths!!!
GENERATED_TEST_DATA = PROJ_ROOT / "tests" / "data_for_tests" / "generated"


def main():
    """loads csv files created by `prep` and changes the parent of paths so it's
    the local file system, instead of what's on my laptop.
    To get tests to run on CI without FileNotFound errors"""
    prep_csvs = sorted(GENERATED_TEST_DATA.glob("**/*prep*csv"))
    for prep_csv in prep_csvs:
        vak_df = pd.read_csv(prep_csv)
        for path_column_name in ("spect_path", "audio_path", "annot_path"):
            paths = vak_df[path_column_name].values.tolist()
            paths = [str(path) for path in paths]
            new_column = []
            for path in paths:
                if path == "nan":
                    continue
                before, aft = path.split("vak/")
                new_path = PROJ_ROOT_ABS.joinpath(aft)
                new_column.append(str(new_path))
            vak_df[path_column_name] = pd.Series(new_column)
        vak_df.to_csv(prep_csv)


if __name__ == "__main__":
    main()



import numpy as np

from .. import files
from ..validators import column_or_1d

from . import functional as F

__all__ = [
    "AddChannel",
    "PadToWindow",
    "StandardizeSpect",
    "ToFloatTensor",
    "ToLongTensor",
    "ViewAsWindowBatch",
]


# adapted from:
# https://github.com/NickleDave/hybrid-vocal-classifier/blob/master/hvc/neuralnet/utils.py
class StandardizeSpect:
    """transform that standardizes spectrograms so they are all
    on the same scale, by subtracting off the mean and dividing by the
    standard deviation from a 'fit' set of spectrograms.

    Attributes
    ----------
    mean_freqs : numpy.ndarray
        mean values for each frequency bin across the fit set of spectrograms
    std_freqs : numpy.ndarray
        standard deviation for each frequency bin across the fit set of spectrograms
    non_zero_std : numpy.ndarray
        boolean, indicates where std_freqs has non-zero values. Used to avoid divide-by-zero errors.
    """

    def __init__(self, mean_freqs=None, std_freqs=None, non_zero_std=None):
        """initialize a new StandardizeSpect instance

        Parameters
        ----------
        mean_freqs : numpy.ndarray
            vector of mean values for each frequency bin across the fit set of spectrograms
        std_freqs : numpy.ndarray
            vector of standard deviations for each frequency bin across the fit set of spectrograms
        non_zero_std : numpy.ndarray
            boolean, indicates where std_freqs has non-zero values. Used to avoid divide-by-zero errors.
        """
        if any([arg is not None for arg in (mean_freqs, std_freqs, non_zero_std)]):
            mean_freqs, std_freqs, non_zero_std = (
                column_or_1d(arr) for arr in (mean_freqs, std_freqs, non_zero_std)
            )
            if (
                len(
                    np.unique(
                        [arg.shape[0] for arg in (mean_freqs, std_freqs, non_zero_std)]
                    )
                )
                != 1
            ):
                raise ValueError(
                    f"mean_freqs, std_freqs, and non_zero_std must all have the same length"
                )

        self.mean_freqs = mean_freqs
        self.std_freqs = std_freqs
        self.non_zero_std = non_zero_std

    @classmethod
    def fit_df(cls, df, spect_key="s"):
        """fits StandardizeSpect instance, given a pandas DataFrame representing a dataset

        Parameters
        ----------
        df : pandas.DataFrame
            loaded from a .csv file representing a dataset, created by vak.io.dataframe.from_files
        spect_key : str
            key in files in 'spect_path' column that maps to spectrograms in arrays.
            Default is 's'.

        Returns
        -------
        standardize_spect : StandardizeSpect
            instance fit to spectrograms in df
        """
        spect_paths = df["spect_path"]
        spect = files.spect.load(spect_paths[0])[spect_key]
        # in files, spectrograms are in orientation (freq bins, time bins)
        # so we take mean and std across columns, i.e. time bins, i.e. axis 1
        mean_freqs = np.mean(spect, axis=1)
        std_freqs = np.std(spect, axis=1)

        for spect_path in spect_paths[1:]:
            spect = files.spect.load(spect_path)[spect_key]
            mean_freqs += np.mean(spect, axis=1)
            std_freqs += np.std(spect, axis=1)
        mean_freqs = mean_freqs / len(spect_paths)
        std_freqs = std_freqs / len(spect_paths)
        non_zero_std = np.argwhere(std_freqs != 0)
        return cls(mean_freqs, std_freqs, non_zero_std)

    @classmethod
    def fit(cls, spect):
        """fit a StandardizeSpect instance.

        Parameters
        ----------
        spect : numpy.ndarray
            with dimensions (frequency bins, time bins)

        Notes
        -----
        Input should be spectrogram.
        Fit function finds the mean and standard deviation of each frequency bin,
        which are used by `transform` method to scale other spectrograms.
        """
        # TODO: make this function accept list and/or ndarray with batch dimension
        if spect.ndim != 2:
            raise ValueError("input spectrogram should be a 2-d array")

        mean_freqs = np.mean(spect, axis=1)
        std_freqs = np.std(spect, axis=1)
        assert mean_freqs.shape[-1] == spect.shape[0]
        assert std_freqs.shape[-1] == spect.shape[0]
        non_zero_std = np.argwhere(std_freqs != 0)
        return cls(mean_freqs, std_freqs, non_zero_std)

    def __call__(self, spect):
        """normalizes input spectrogram with fit parameters.

        Parameters
        ----------
        spect : numpy.ndarray
            2-d array with dimensions (frequency bins, time bins).

        Returns
        -------
        z_norm_spect : numpy.ndarray
            array standardized to same scale as set of spectrograms that
            SpectScaler was fit with
        """
        if any([not hasattr(self, attr) for attr in ["mean_freqs", "std_freqs"]]):
            raise AttributeError(
                "SpectScaler properties are set to None,"
                "must call fit method first to set the"
                "value of these properties before calling"
                "transform"
            )

        if type(spect) != np.ndarray:
            raise TypeError(
                f"type of spect must be numpy.ndarray but was: {type(spect)}"
            )

        if spect.shape[0] != self.mean_freqs.shape[0]:
            raise ValueError(
                f"number of rows in spects, {spect.shape[0]}, "
                f"does not match number of elements in self.mean_freqs, {self.mean_freqs.shape[0]},"
                "i.e. the number of frequency bins from the spectrogram"
                "to which the scaler was fit originally"
            )

        return F.standardize_spect(
            spect, self.mean_freqs, self.std_freqs, self.non_zero_std
        )

    def __repr__(self):
        args = f"(mean_freqs={self.mean_freqs}, std_freqs={self.std_freqs}, non_zero_std={self.non_zero_std})"
        return self.__class__.__name__ + args


class PadToWindow:
    """pad a 1d or 2d array so that it can be reshaped
    into consecutive windows of specified size

    Parameters
    ----------
    arr : numpy.ndarray
        with 1 or 2 dimensions, e.g. a vector of labeled timebins
        or a spectrogram.
    window_size : int
        width of window in number of elements.
    padval : float
        value to pad with. Added to end of array, the
        "right side" if 2-dimensional.
    return_padding_mask : bool
        if True, return a boolean vector to use for cropping
        back down to size before padding. padding_mask has size
        equal to width of padded array, i.e. original size
        plus padding at the end, and has values of 1 where
        columns in padded are from the original array,
        and values of 0 where columns were added for padding.

    Returns
    -------
    padded : numpy.ndarray
        padded with padval
    padding_mask : np.bool
        has size equal to width of padded, i.e. original size
        plus padding at the end. Has values of 1 where
        columns in padded are from the original array,
        and values of 0 where columns were added for padding.
        Only returned if return_padding_mask is True.
    """

    def __init__(self, window_size, padval=0.0, return_padding_mask=True):
        if not (type(window_size) == int) or (
            type(window_size) == float and window_size.is_integer() is False
        ):
            raise ValueError(
                f"window size must be an int or a whole number float;"
                f" type was {type(window_size)} and value was {window_size}"
            )

        if type(padval) not in (int, float):
            raise TypeError(
                f"type for padval must be int or float but was: {type(padval)}"
            )
        if not type(return_padding_mask) == bool:
            raise TypeError(
                "return_padding_mask must be boolean (True or False), "
                f"but was type {type(return_padding_mask)} with value {return_padding_mask}"
            )

        self.window_size = window_size
        self.padval = padval
        self.return_padding_mask = return_padding_mask

    def __call__(self, arr):
        return F.pad_to_window(
            arr, self.window_size, self.padval, self.return_padding_mask
        )

    def __repr__(self):
        args = f"(window_size={self.window_size}, padval={self.padval}, return_padding_mask={self.return_padding_mask})"
        return self.__class__.__name__ + args


class ViewAsWindowBatch:
    """return view of a 1d or 2d array as a batch of non-overlapping windows

    Parameters
    ----------
    arr : numpy.ndarray
        with 1 or 2 dimensions, e.g. a vector of labeled timebins
        or a 2-d array representing a spectrogram.
        If the array has 2-d dimensions, the returned array will
        have dimensions (batch, height of array, window width)
    window_width : int
        width of window in number of elements.

    Returns
    -------
    batch_windows : numpy.ndarray
        with shape (batch size, window_width) if array is 1d,
        or with shape (batch size, height, window_width) if array is 2d.
        Batch size will be arr.shape[-1] // window_width.
        Window width must divide arr.shape[-1] evenly.
        To pad the array so it can be divided into windows of the specified
        width, use the `pad_to_window` transform

    Notes
    -----
    adapted from skimage.util.view_as_blocks
    https://github.com/scikit-image/scikit-image/blob/f1b7cf60fb80822849129cb76269b75b8ef18db1/skimage/util/shape.py#L9
    """

    def __init__(self, window_width):
        if not (type(window_width) == int) or (
            type(window_width) == float and window_width.is_integer() is False
        ):
            raise ValueError(
                f"window size must be an int or a whole number float;"
                f" type was {type(window_width)} and value was {window_width}"
            )

        self.window_width = window_width

    def __call__(self, arr):
        return F.view_as_window_batch(arr, self.window_width)

    def __repr__(self):
        args = f"(window_width={self.window_width})"
        return self.__class__.__name__ + args


class ToFloatTensor:
    """convert Numpy array to torch.FloatTensor.

    Parameters
    ----------
    arr : numpy.ndarray

    Returns
    -------
    float_tensor
        with dtype 'float32'
    """

    def __init__(self):
        pass

    def __call__(self, arr):
        return F.to_floattensor(arr)

    def __repr__(self):
        return self.__class__.__name__


class ToLongTensor:
    """convert Numpy array to torch.LongTensor.

    Parameters
    ----------
    arr : numpy.ndarray

    Returns
    -------
    long_tensor : torch.Tensor
        with dtype 'float64'
    """

    def __init__(self):
        pass

    def __call__(self, arr):
        return F.to_longtensor(arr)

    def __repr__(self):
        return self.__class__.__name__


class AddChannel:
    """add a channel dimension to a 2-dimensional tensor.
    Transform that makes it easy to treat a spectrogram as an image,
    by adding a dimension with a single 'channel', analogous to grayscale.
    In this way the tensor can be fed to e.g. convolutional layers.

    Parameters
    ----------
    input : torch.Tensor
        with two dimensions (height, width).
    channel_dim : int
        dimension where "channel" is added.
        Default is 0, which returns a tensor with dimensions (channel, height, width).
    """

    def __init__(self, channel_dim=0):
        if not (type(channel_dim) == int) or (
            type(channel_dim) == float and channel_dim.is_integer() is False
        ):
            raise ValueError(
                f"window size must be an int or a whole number float;"
                f" type was {type(channel_dim)} and value was {channel_dim}"
            )

        channel_dim = int(channel_dim)

        if channel_dim < 0 and channel_dim != -1:
            raise ValueError(
                "value of channel_dim should be a non-negative integer, or -1 (for last dimension). "
                f"Value was: {channel_dim}"
            )

        self.channel_dim = channel_dim

    def __call__(self, input):
        return F.add_channel(input, channel_dim=self.channel_dim)

    def __repr__(self):
        args = f"(channel_dim={self.channel_dim})"
        return self.__class__.__name__ + args



from pathlib import Path

import numpy as np
from dask import bag as db
from dask.diagnostics import ProgressBar

from .. import constants
from ..logging import log_or_print
from .files import find_fname
from ..timebins import timebin_dur_from_vec


def find_audio_fname(spect_path, audio_ext=None):
    """finds name of audio file in a path to a spectrogram file,
    if one is present.

    Checks for any extension that is a valid audio file format
    and returns path up to and including that extension,
    i.e. with the spectrogram file extension removed.

    Parameters
    ----------
    spect_path : str, Path
        path to a spectrogram file
    audio_ext : str
        extension associated with an audio file format, used to
        find audio file name in spect_path.
        Default is None. If None, search for any valid audio format
        (as defined by vak.config.constants.VALID_AUDIO_FORMATS)

    Returns
    -------
    audio_fname : str
        name of audio file found in spect_path
    """
    if audio_ext is None:
        audio_ext = constants.VALID_AUDIO_FORMATS
    elif type(audio_ext) is str:
        audio_ext = [audio_ext]
    else:
        raise TypeError(f"invalid type for audio_ext: {type(audio_ext)}")

    audio_fnames = []
    for ext in audio_ext:
        audio_fnames.append(find_fname(str(spect_path), ext))
    # remove Nones
    audio_fnames = [path for path in audio_fnames if path is not None]
    # keep just file name from spect path
    audio_fnames = [Path(path).name for path in audio_fnames]

    if len(audio_fnames) == 1:
        return audio_fnames[0]
    else:
        raise ValueError(
            f"unable to determine filename of audio file from: {spect_path}"
        )


def load(spect_path, spect_format=None):
    """load spectrogram and related arrays from a file,
    return as an object that provides Python dictionary-like
    access

    Parameters
    ----------
    spect_path : str, Path
        to an array file.
    spect_format : str
        Valid formats are defined in vak.io.spect.SPECT_FORMAT_LOAD_FUNCTION_MAP.
        Default is None, in which case the extension of the file is used.

    Returns
    -------
    spect_dict : dict-like
        either a dictionary or dictionary-like object that provides access to arrays
        from the file via keys, e.g. spect_dict['s'] for the spectrogram.
        See docstring for vak.audio.to_spect for default keys for spectrogram
        array files that function creates.
    """
    spect_path = Path(spect_path)
    if spect_format is None:
        # "replace('.', '')", because suffix returns file extension with period included
        spect_format = spect_path.suffix.replace(".", "")
    spect_dict = constants.SPECT_FORMAT_LOAD_FUNCTION_MAP[spect_format](spect_path)
    return spect_dict


def timebin_dur(spect_path, spect_format, timebins_key, n_decimals_trunc=5):
    """get duration of time bins from a spectrogram file

    Parameters
    ----------
    spect_path: str, Path
        path to spectrogram file.
    spect_format : str
        format of file containing spectrogram. One of {'mat', 'npz'}
    timebins_key : str
        key for accessing vector of time bins in files. Default is 't'.
    n_decimals_trunc : int
        number of decimal places to keep when truncating the timebin duration calculated from
        the vector of time bins.
        Default is 3, i.e. assumes milliseconds is the last significant digit.

    Returns
    -------
    timebin_dur : float

    """
    spect_path = Path(spect_path)
    spect_dict = load(spect_path, spect_format)
    time_bins = spect_dict[timebins_key]
    timebin_dur = timebin_dur_from_vec(time_bins, n_decimals_trunc)
    return timebin_dur


def is_valid_set_of_spect_files(
    spect_paths,
    spect_format,
    freqbins_key="f",
    timebins_key="t",
    spect_key="s",
    n_decimals_trunc=5,
    logger=None,
):
    """validate a set of spectrogram files that will be used as a dataset.
    Validates that:
      - all files contain a spectrogram array that can be accessed with the specified key
      - the length of the frequency bin array in each file equals the number of rows in the spectrogram array
      - the frequency bins are the same across all files
      - the length of the time bin array in each file equals the number of columns in the spectrogram array
      - the duration of a spectrogram time bin is the same across all files

    Parameters
    ----------
    spect_paths: list
        of strings or pathlib.Path objects; paths to spectrogram files.
    spect_format : str
        format of files containing spectrograms. One of {'mat', 'npz'}
    freqbins_key : str
        key for accessing vector of frequency bins in files. Default is 'f'.
    timebins_key : str
        key for accessing vector of time bins in files. Default is 't'.
    spect_key : str
        key for accessing spectrogram in files. Default is 's'.
    n_decimals_trunc : int
        number of decimal places to keep when truncating the timebin duration calculated from
        the vector of time bins.
        Default is 3, i.e. assumes milliseconds is the last significant digit.

    Other Parameters
    ----------------
    logger : logging.Logger
        instance created by vak.logging.get_logger. Default is None.

    Returns
    -------
    returns True if all validation checks pass. If not, an error is raised.
    """
    spect_paths = [Path(spect_path) for spect_path in spect_paths]

    def _validate(spect_path):
        """validates each spectrogram file, then returns frequency bin array
        and duration of time bins, so that those can be validated across all files"""
        spect_dict = load(spect_path, spect_format)

        if spect_key not in spect_dict:
            raise KeyError(
                f"Did not find a spectrogram in file '{spect_path.name}' "
                f"using spect_key '{spect_key}'."
            )

        freq_bins = spect_dict[freqbins_key]
        time_bins = spect_dict[timebins_key]
        timebin_dur = timebin_dur_from_vec(time_bins, n_decimals_trunc)

        # number of freq. bins should equal number of rows
        if spect_dict[freqbins_key].shape[-1] != spect_dict[spect_key].shape[0]:
            raise ValueError(
                f"length of frequency bins in {spect_path.name} "
                "does not match number of rows in spectrogram"
            )
        # number of time bins should equal number of columns
        if spect_dict[timebins_key].shape[-1] != spect_dict[spect_key].shape[1]:
            raise ValueError(
                f"length of time_bins in {spect_path.name} "
                f"does not match number of columns in spectrogram"
            )

        return spect_path, freq_bins, timebin_dur

    spect_paths_bag = db.from_sequence(spect_paths)

    log_or_print("validating set of spectrogram files", logger=logger, level="info")

    with ProgressBar():
        path_freqbins_timebin_dur_tups = list(spect_paths_bag.map(_validate))

    all_freq_bins = np.stack([tup[1] for tup in path_freqbins_timebin_dur_tups])
    uniq_freq_bins = np.unique(all_freq_bins, axis=0)
    if len(uniq_freq_bins) != 1:
        raise ValueError(
            f"Found more than one frequency bin vector across files. "
            f"Instead found {len(uniq_freq_bins)}"
        )

    timebin_durs = [tup[2] for tup in path_freqbins_timebin_dur_tups]
    uniq_durs = np.unique(timebin_durs)
    if len(uniq_durs) != 1:
        raise ValueError(
            "Found more than one duration for time bins across spectrogram files. "
            f"Durations found were: {uniq_durs}"
        )

    return True



from collections import defaultdict
import pprint
import re
import shutil

import numpy as np
import pandas as pd

from ... import split
from ...converters import expanded_user_path
from ...datasets.window_dataset import WindowDataset
from ...logging import log_or_print


# pattern used by path.glob to find all training data subset csvs within previous_run_path
CSV_GLOB = "**/*prep*csv"
# pattern used by re to get training subset duration
TRAIN_DUR_PAT = r"train_dur_(\d+\.\d+|\d+)s"


def _dict_from_dir(previous_run_path):
    """
    build dictionary that maps training set durations to a list of
    training subset csv paths, ordered by replicate number

    factored out as helper function so we can test this works correctly

    Parameters
    ----------
    previous_run_path : str, Path
        path to directory containing dataset .csv files
        that represent subsets of training set, created by
        a previous run of ``vak.core.learncurve.learning_curve``.
        Typically directory will have a name like ``results_{timestamp}``
        and the actual .csv splits will be in sub-directories with names
        corresponding to the training set duration

    Returns
    -------
    train_dur_csv_paths : dict
        where keys are duration in seconds of subsets taken from training data,
        and corresponding values are lists of paths to .csv files containing
        those subsets
    """
    train_dur_csv_paths = {}
    train_dur_dirs = previous_run_path.glob("train_dur_*s")
    for train_dur_dir in train_dur_dirs:
        train_dur = re.findall(TRAIN_DUR_PAT, train_dur_dir.name)
        if len(train_dur) != 1:
            raise ValueError(
                f"did not find just a single training subset duration in filename:\n"
                f"{train_subset_path}\n"
                f"Instead found: {train_dur}"
            )
        train_dur = int(train_dur[0])
        # sort by increasing replicate number -- numerically, not alphabetically
        replicate_dirs = sorted(
            train_dur_dir.glob("replicate_*"),
            key=lambda dir_path: int(dir_path.name.split("_")[-1]),
        )
        train_subset_paths = []
        for replicate_dir in replicate_dirs:
            train_subset_path = sorted(replicate_dir.glob("*prep*csv"))
            if len(train_subset_path) != 1:
                raise ValueError(
                    f"did not find just a single training subset .csv in replicate directory:\n"
                    f"{replicate_dir}\n"
                    f"Instead found: {train_subset_path}"
                )
            train_subset_path = train_subset_path[0]
            train_subset_paths.append(train_subset_path)
        train_dur_csv_paths[train_dur] = train_subset_paths

    return train_dur_csv_paths


def from_dir(
    previous_run_path,
    train_set_durs,
    timebin_dur,
    num_replicates,
    results_path,
    window_size,
    spect_key,
    timebins_key,
    labelmap,
    logger=None,
):
    """return a ``dict`` mapping training dataset durations to dataset csv paths
    from a previous run of `vak.core.learncurve.learning_curve`.

    Recovers the previous subsets of the total training set, so that
    they can be used to re-run a different experiment on the same subsets for comparison.

    Parameters
    ----------
    previous_run_path : str, Path
        path to directory containing dataset .csv files
        that represent subsets of training set, created by
        a previous run of ``vak.core.learncurve.learning_curve``.
        Typically directory will have a name like ``results_{timestamp}``
        and the actual .csv splits will be in sub-directories with names
        corresponding to the training set duration
    train_set_durs : list
        of int, durations in seconds of subsets taken from training data
        to create a learning curve, e.g. [5, 10, 15, 20].
    timebin_dur : float
        duration of timebins in spectrograms
    num_replicates : int
        number of times to replicate training for each training set duration
        to better estimate metrics for a training set of that size.
        Each replicate uses a different randomly drawn subset of the training
        data (but of the same duration).
    results_path : str, pathlib.Path
        Directory where results will be saved, including
        files representing subsets of training data that this function makes.
        Path derived from the ``root_results_dir`` argument
         to ``vak.core.learncurve.learning_curve``, unless specified by user.
    window_size : int
        size of windows taken from spectrograms, in number of time bins,
        shown to neural networks
    spect_key : str
        key for accessing spectrogram in files. Default is 's'.
    timebins_key : str
        key for accessing vector of time bins in files. Default is 't'.
    labelmap : dict
        that maps labelset to consecutive integers

    Other Parameters
    ----------------
    logger : logging.Logger
        instance created by ``vak.logging.get_logger``. Default is None.

    Returns
    -------
    train_dur_csv_paths : dict
        where keys are duration in seconds of subsets taken from training data,
        and corresponding values are lists of paths to .csv files containing
        those subsets

    Notes
    -----
    ``train_set_durs`` and ``num_replicates`` are used to verify that
    the correct number of training data subsets are found for each specified
    duration, i.e., that the new experiment will correctly use
    all datasets from a previous run
    """
    previous_run_path = expanded_user_path(previous_run_path)
    if not previous_run_path.is_dir():
        raise NotADirectoryError(
            f"previous_run_path not recognized as a directory:\n{previous_run_path}"
        )

    train_dur_csv_paths = _dict_from_dir(previous_run_path)

    # validate results
    found_train_set_durs = sorted(train_dur_csv_paths.keys())
    if not found_train_set_durs == sorted(train_set_durs):
        raise ValueError(
            f"training set durations found in {previous_run_path} did not match "
            f"specified training set durations.\n"
            f"Specified: {sorted(train_set_durs)}\n"
            f"Found: {found_train_set_durs}"
        )

    if not all(
        [len(csv_paths) == num_replicates for csv_paths in train_dur_csv_paths.values()]
    ):
        raise ValueError(
            "did not find correct number of replicates for all training set durations."
            f"Found the following:\n{pprint.pformat(train_dur_csv_paths, indent=4)}"
        )

    log_or_print(
        f"Using the following training subsets from previous run path:\n"
        f"{pprint.pformat(train_dur_csv_paths, indent=4)}",
        logger=logger,
        level="info",
    )

    # need to copy .csv files, and change path in train_dur_csv_paths to point to copies
    # so that `vak.train` doesn't try to write over existing results dir, causing a crash
    for (
        train_dur
    ) in train_dur_csv_paths.keys():  # use keys so we can modify dict inside loop
        results_path_this_train_dur = results_path.joinpath(f"train_dur_{train_dur}s")
        results_path_this_train_dur.mkdir()
        csv_paths = train_dur_csv_paths[train_dur]
        new_csv_paths = []
        for replicate_num, csv_path in zip(range(1, len(csv_paths) + 1), csv_paths):
            results_path_this_replicate = results_path_this_train_dur.joinpath(
                f"replicate_{replicate_num}"
            )
            results_path_this_replicate.mkdir()
            # copy csv using Path.rename() method, append returned new path to list
            new_csv_paths.append(
                shutil.copy(
                    src=csv_path,
                    dst=results_path_this_replicate.joinpath(csv_path.name),
                )
            )

            subset_df = pd.read_csv(csv_path)
            (
                spect_id_vector,
                spect_inds_vector,
                x_inds,
            ) = WindowDataset.spect_vectors_from_df(
                subset_df,
                "train",
                window_size,
                spect_key,
                timebins_key,
                crop_dur=train_dur,
                timebin_dur=timebin_dur,
                labelmap=labelmap,
            )
            for vec_name, vec in zip(
                ["spect_id_vector", "spect_inds_vector", "x_inds"],
                [spect_id_vector, spect_inds_vector, x_inds],
            ):
                np.save(results_path_this_replicate.joinpath(f"{vec_name}.npy"), vec)

        train_dur_csv_paths[train_dur] = new_csv_paths

    return train_dur_csv_paths


def train_dur_dirname(train_dur):
    """helper function that returns name of directory for all replicates
    trained with a training set of a specified duration

    factored out as function so we can test and use in fixtures
    """
    return f"train_dur_{train_dur}s"


def replicate_dirname(replicate_num):
    """ "helper function that returns name of directory for a replicate

    factored out as function so we can test and use in fixtures
    """
    return f"replicate_{replicate_num}"


def subset_csv_filename(csv_path, train_dur, replicate_num):
    """ "helper function that returns name of directory for a replicate

    factored out as function so we can test and use in fixtures
    """
    return f"{csv_path.stem}_train_dur_{train_dur}s_replicate_{replicate_num}.csv"


def from_df(
    dataset_df,
    csv_path,
    train_set_durs,
    timebin_dur,
    num_replicates,
    results_path,
    labelset,
    window_size,
    spect_key,
    timebins_key,
    labelmap,
    logger=None,
):
    """return a ``dict`` mapping training dataset durations to dataset csv paths.

    csv paths representing subsets of the training data are generated using ``vak.split``.

    Parameters
    ----------
    dataset_df : pandas.DataFrame
        representing an entire dataset of vocalizations.
    csv_path : pathlib.Path
        path to where dataset was saved as a csv.
    train_set_durs : list
        of int, durations in seconds of subsets taken from training data
        to create a learning curve, e.g. [5, 10, 15, 20].
    timebin_dur : float
        duration of timebins in spectrograms
    num_replicates : int
        number of times to replicate training for each training set duration
        to better estimate metrics for a training set of that size.
        Each replicate uses a different randomly drawn subset of the training
        data (but of the same duration).
    results_path : str, pathlib.Path
        Directory where results will be saved, including
        files representing subsets of training data that this function makes.
        Path derived from the ``root_results_dir`` argument
         to ``vak.core.learncurve.learning_curve``, unless specified by user.
    labelset : set
        of str or int, the set of labels that correspond to annotated segments
        that a network should learn to segment and classify. Note that if there
        are segments that are not annotated, e.g. silent gaps between songbird
        syllables, then `vak` will assign a dummy label to those segments
        -- you don't have to give them a label here.
    window_size : int
        size of windows taken from spectrograms, in number of time bins,
        shonw to neural networks
    spect_key : str
        key for accessing spectrogram in files. Default is 's'.
    timebins_key : str
        key for accessing vector of time bins in files. Default is 't'.
    labelmap : dict
        that maps labelset to consecutive integers

    Other Parameters
    ----------------
    logger : logging.Logger
        instance created by ``vak.logging.get_logger``. Default is None.

    Returns
    -------
    train_dur_csv_paths : dict
        where keys are duration in seconds of subsets taken from training data,
        and corresponding values are lists of paths to .csv files containing
        those subsets
    """
    train_dur_csv_paths = defaultdict(list)
    for train_dur in train_set_durs:
        log_or_print(
            f"subsetting training set for training set of duration: {train_dur}",
            logger=logger,
            level="info",
        )
        results_path_this_train_dur = results_path.joinpath(
            train_dur_dirname(train_dur)
        )
        results_path_this_train_dur.mkdir()
        for replicate_num in range(1, num_replicates + 1):
            results_path_this_replicate = results_path_this_train_dur.joinpath(
                replicate_dirname(replicate_num)
            )
            results_path_this_replicate.mkdir()
            # get just train split, to pass to split.dataframe
            # so we don't end up with other splits in the training set
            train_split_df = dataset_df[dataset_df["split"] == "train"]
            train_split_df = split.dataframe(
                train_split_df, train_dur=train_dur, labelset=labelset
            )
            train_split_df = train_split_df[
                train_split_df.split == "train"
            ]  # remove rows where split set to 'None'
            # ---- use *just* train subset to get spect vectors for WindowDataset
            (
                spect_id_vector,
                spect_inds_vector,
                x_inds,
            ) = WindowDataset.spect_vectors_from_df(
                train_split_df,
                "train",
                window_size,
                spect_key,
                timebins_key,
                crop_dur=train_dur,
                timebin_dur=timebin_dur,
                labelmap=labelmap,
            )
            for vec_name, vec in zip(
                ["spect_id_vector", "spect_inds_vector", "x_inds"],
                [spect_id_vector, spect_inds_vector, x_inds],
            ):
                np.save(results_path_this_replicate.joinpath(f"{vec_name}.npy"), vec)
            # keep the same validation and test set by concatenating them with the train subset
            subset_df = pd.concat(
                (
                    train_split_df,
                    dataset_df[dataset_df.split == "val"],
                    dataset_df[dataset_df.split == "test"],
                )
            )

            subset_csv_path = results_path_this_replicate.joinpath(
                subset_csv_filename(csv_path, train_dur, replicate_num)
            )
            subset_df.to_csv(subset_csv_path, index=False)
            train_dur_csv_paths[train_dur].append(subset_csv_path)

    return train_dur_csv_paths



"""parses [LEARNCURVE] section of config"""
import attr
from attr import converters, validators
from attr.validators import instance_of

from .validators import is_a_directory
from .train import TrainConfig
from ..converters import expanded_user_path


@attr.s
class LearncurveConfig(TrainConfig):
    """class that represents [LEARNCURVE] section of config.toml file

    Attributes
    ----------
    models : list
        of model names. e.g., 'models = TweetyNet, GRUNet, ConvNet'
    csv_path : str
        path to where dataset was saved as a csv.
    num_epochs : int
        number of training epochs. One epoch = one iteration through the entire
        training set.
    normalize_spectrograms : bool
        if True, use spect.utils.data.SpectScaler to normalize the spectrograms.
        Normalization is done by subtracting off the mean for each frequency bin
        of the training set and then dividing by the std for that frequency bin.
        This same normalization is then applied to validation + test data.
    ckpt_step : int
        step/epoch at which to save to checkpoint file.
        Default is None, in which case checkpoint is only saved at the last epoch.
    patience : int
        number of epochs to wait without the error dropping before stopping the
        training. Default is None, in which case training continues for num_epochs
    train_set_durs : list
        of int, durations in seconds of subsets taken from training data
        to create a learning curve, e.g. [5, 10, 15, 20]. Default is None
        (when training a single model on all available training data).
    num_replicates : int
        number of times to replicate training for each training set duration
        to better estimate mean accuracy for a training set of that size.
        Each replicate uses a different randomly drawn subset of the training
        data (but of the same duration).
    save_only_single_checkpoint_file : bool
        if True, save only one checkpoint file instead of separate files every time
        we save. Default is True.
    use_train_subsets_from_previous_run : bool
        if True, use training subsets saved in a previous run. Default is False.
        Requires setting previous_run_path option in config.toml file.
    previous_run_path : str
        path to results directory from a previous run.
        Used for training if use_train_subsets_from_previous_run is True.
    """

    train_set_durs = attr.ib(validator=instance_of(list), kw_only=True)
    num_replicates = attr.ib(validator=instance_of(int), kw_only=True)
    previous_run_path = attr.ib(
        converter=converters.optional(expanded_user_path),
        validator=validators.optional(is_a_directory),
        default=None,
    )



import numpy as np

import vak.timebins


def test_timebin_dur_from_vecs():
    timebin_dur = 0.001
    time_bins = np.linspace(0.0, 5.0, num=int(5 / timebin_dur))
    computed = vak.timebins.timebin_dur_from_vec(
        time_bins=time_bins, n_decimals_trunc=3
    )
    assert timebin_dur == computed



import logging
import random

from .validate import validate_split_durations


def brute_force(durs, labels, labelset, train_dur, val_dur, test_dur, max_iter=5000):
    """finds indices that split (labels, durations) tuples into training,
    test, and validation sets of specified durations, with the set of unique labels
    in each dataset equal to the specified labelset.

    The durations of the datasets created using the returned indices will be
    *greater than* or equal to the durations specified.

    Must specify a positive value for one of {train_dur, test_dur}.
    The other value can be specified as '-1' which is interpreted as
    "use the remainder of the dataset for this split,
    after finding indices for the set with a specified duration".

    Parameters
    ----------
    durs : list
        of durations of vocalizations
    labels : list
        of labels from vocalizations
    labelset : set
        of labels
    train_dur : int, float
        Target duration for training set, in seconds.
    val_dur : int, float
        Target duration for validation set, in seconds.
    test_dur : int, float
        Target duration for test set, in seconds.
    max_iter : int
        maximum number of iterations to attempt to find indices. Default is 5000.

    Returns
    -------
    train_inds, val_inds, test_inds : list
        of int, the indices that will split datasets

    Notes
    -----
    A 'brute force' algorithm that just randomly assigns indices to a set,
    and iterates until it finds some partition where each set has instances of all classes of label.
    Starts by ensuring that each label is represented in each set and then adds files to reach the required
    durations.
    """
    logger = logging.getLogger(__name__)
    logger.setLevel("INFO")

    sum_durs = sum(durs)
    train_dur, val_dur, test_dur = validate_split_durations(
        train_dur, val_dur, test_dur, sum_durs
    )
    target_split_durs = dict(
        zip(("train", "val", "test"), (train_dur, val_dur, test_dur))
    )

    if not len(durs) == len(labels):
        raise ValueError(
            "length of list of durations did not equal length of list of labels; "
            "should be same length since "
            "each duration of a vocalization corresponds to the labels from its annotations.\n"
            f"Length of durations: {len(durs)}. Length of labels: {len(labels)}"
        )

    iter = 1
    all_labels_err = (
        "Did not successfully divide data into training, "
        "validation, and test sets of sufficient duration "
        f"after {max_iter} iterations. "
        "Try increasing the total size of the data set."
    )

    # ---- outer loop that repeats until we successfully split our reach max number of iters ---------------------------
    while 1:
        # list of indices we use to index into both `durs` and `labels`
        durs_labels_inds = list(
            range(len(labels))
        )  # we checked len(labels) == len(durs) above

        # when making `split_inds`, "initialize" the dict with all split names, by using target_split_durs
        # so we don't get an error when indexing into dict in return statement below
        split_inds = {split_name: [] for split_name in target_split_durs.keys()}
        total_split_durs = {split_name: 0 for split_name in target_split_durs.keys()}
        split_labelsets = {split_name: set() for split_name in target_split_durs.keys()}

        # list of split 'choices' we use when randomly adding indices to splits
        choice = []
        for split_name in target_split_durs.keys():
            if target_split_durs[split_name] > 0 or target_split_durs[split_name] == -1:
                choice.append(split_name)

        # ---- make sure each split has at least one instance of each label --------------------------------------------
        for label_from_labelset in sorted(labelset):
            label_inds = [
                ind for ind in durs_labels_inds if label_from_labelset in labels[ind]
            ]

            random.shuffle(label_inds)
            for split_name in target_split_durs.keys():
                if (
                    target_split_durs[split_name] > 0
                    or target_split_durs[split_name] == -1
                ) and label_from_labelset not in split_labelsets[split_name]:
                    try:
                        ind = label_inds.pop()
                        split_inds[split_name].append(ind)
                        total_split_durs[split_name] += durs[ind]
                        split_labelsets[split_name] = split_labelsets[split_name].union(
                            set(labels[ind])
                        )
                        durs_labels_inds.remove(ind)
                    except IndexError:
                        if len(label_inds) == 0:
                            logger.debug(
                                "Ran out of elements while dividing dataset into subsets of specified durations."
                                f"Iteration {iter}"
                            )
                            iter += 1
                            break  # do next iteration
                        else:
                            # something else happened, re-raise error
                            raise

        for split_name in target_split_durs.keys():
            if (
                target_split_durs[split_name] > 0
                and total_split_durs[split_name] >= target_split_durs[split_name]
            ):
                choice.remove(split_name)

        if len(choice) == 0:
            finished = True
        else:
            finished = False

        # ---- inner loop that actually does split ---------------------------------------------------------------------
        random.shuffle(durs_labels_inds)
        while finished is False:
            # pop durations off list and append to randomly-chosen
            # list, either train, val, or test set.
            # Do this until the total duration for each data set is equal
            # to or greater than the target duration for each set.
            try:
                ind = durs_labels_inds.pop()
            except IndexError:
                if len(durs_labels_inds) == 0:
                    logger.debug(
                        "Ran out of elements while dividing dataset into subsets of specified durations."
                        f"Iteration {iter}"
                    )
                    iter += 1
                    break  # do next iteration
                else:
                    # something else happened, re-raise error
                    raise

            which_set = random.randint(0, len(choice) - 1)
            split_name = choice[which_set]
            split_inds[split_name].append(ind)
            total_split_durs[split_name] += durs[ind]
            if (
                target_split_durs[split_name] > 0
                and total_split_durs[split_name] >= target_split_durs[split_name]
            ):
                choice.remove(split_name)
            elif target_split_durs[split_name] == -1:
                # if this split is -1 and other split is already "finished"
                if (split_name == "test" and "train" not in choice) or (
                    split_name == "train" and "test" not in choice
                ):
                    # just add all remaining inds to this split
                    split_inds[split_name].extend(durs_labels_inds)
                    choice.remove(split_name)

            if len(choice) < 1:  # list is empty, we popped off all the choices
                for split_name in target_split_durs.keys():
                    if target_split_durs[split_name] > 0:
                        if total_split_durs[split_name] < target_split_durs[split_name]:
                            raise ValueError(
                                "Loop to find splits completed, "
                                f"but total duration of '{split_name}' split, "
                                f"{total_split_durs[split_name]} seconds, "
                                f"is less than target duration specified: {target_split_durs[split_name]} seconds."
                            )
                else:
                    finished = True
                    break

        if iter > max_iter:
            raise ValueError(
                "Could not find subsets of sufficient duration in "
                f"less than {max_iter} iterations."
            )

        # make sure that each split contains all unique labels in labelset
        if finished is True:
            for split_name in target_split_durs.keys():
                if (
                    target_split_durs[split_name] > 0
                    or target_split_durs[split_name] == -1
                ):
                    split_labels = [
                        label for ind in split_inds[split_name] for label in labels[ind]
                    ]
                    split_labelset = set(split_labels)
                    if split_labelset != set(labelset):
                        iter += 1
                        if iter > max_iter:
                            raise ValueError(all_labels_err)
                        else:
                            logger.debug(
                                f"Set of unique labels in '{split_name}' split did not equal specified labelset. "
                                f"Getting new '{split_name}' split. Iteration: {iter}"
                            )
                            continue

            # successfully split
            break

        elif finished is False:
            continue

    split_inds = {
        split_name: (inds if inds else None) for split_name, inds in split_inds.items()
    }

    return split_inds["train"], split_inds["val"], split_inds["test"]



"""script run by Makefile test-data-generate command

makes all the 'generated' test data, i.e. files created by vak,
It's called 'generated' test data to distinguish it from the
'source' test data, i.e., files **not** created by vak, that is,
the input data used when vak does create files (csv files, logs,
neural network checkpoints, etc.)

This script generates:
* temporary config.toml files used when generating results
* `prep`d (prepared) datasets, and results created with those datasets,
  both of which were generated using the temporary config.toml files

all the setup configs send output to one of two places:
for any prep command, the output goes to some child directory of ./tests/data_for_tests/generated/prep
for any command run with a `prep`d dataset, the output goes to some child dir of ./tests/data_for_tests/generated/results

examples:
    when we run `vak prep tests/data_for_tests/generated/configs/test_train_audio_wav_annot_birdsongrec.toml`
    the `prep`d dataset will be in a new directory created in
    `./tests/data_for_tests/generated/prep/train/audio_wav_annot_birdsongrec`

    when we run `vak train tests/data_for_tests/genereated/configs/test_train_audio_wav_annot_birdsongrec.toml`
    it will use the `prep`d dataset csv that's now in
    `./tests/data_for_tests/generated/prep/train/audio_wav_annot_birdsongrec`, that the temporary config.toml points to,
    and the results will go to a new directory created in
    `./tests/data_for_tests/generated/results/train/audio_wav_annot_birdsongrec`

To set up this directory structure, we iterate through the constants defined below.

The constants are:
- TOP_LEVEL_DIRS
    name of sub-directories in .tests/data_for_tests/generated that correspond to
    either tempory config files, 'prep'd datasets, or `results` generated from those `prep`d datasets
- COMMAND_DIRS
    names of sub-sub-directories in ./tests/data_for_tests/generated/prep &
    ./tests/data_for_tests/generated/results that correspond to cli commands;
    e.g., dataset from running `vak prep $A_TRAIN_CONFIG.TOML` will be in
    ./tests/data_for_tests/generated/prep/train/audio_{}_annot_{}
- DATA_DIRS
    names of "subsub"directories in ./tests/data_for_tests/$COMMAND that correspond to toy test data sets;
    those sub-directories contain test data generated by $COMMAND using the specified toy test data set

In other words, the parent directory for all the generated directories we need to remove
will have a path of the form: `./tests/data_for_tests/$TOP_LEVEL_DIRS/COMMAND_DIRS/DATA_DIRS`.
For example:
The prep`d dataset from running `vak prep $AUDIO_CBIN_ANNOT_NOTMAT_CONFIG.TOML` will be in
`./tests/data_for_tests/prep/train/audio_cbin_annot_notmat`
and corresponding results will be in
`./tests/data_for_tests/results/train/audio_cbin_annot_notmat`
The directories will have names with timestamps like `prep_20201015_1115`.
Those are the generated directories we want to remove.
"""
from pathlib import Path
import shutil

import toml
import vak

HERE = Path(__file__).parent
TEST_DATA_ROOT = HERE / ".." / "data_for_tests"
GENERATED_TEST_DATA = TEST_DATA_ROOT / "generated"
GENERATED_TEST_CONFIGS_ROOT = GENERATED_TEST_DATA / "configs"

# convention is that all the config.toml files in tests/data_for_tests/configs
# that should be run when generating test data
# have filenames of the form `{MODEL}_{COMMAND}_audio_{FORMAT}_annot_{FORMAT}.toml'
# **or** `{MODEL}_{COMMAND}_spect_{FORMAT}_annot_{FORMAT}_config.ini'
# e.g., 'tweetynet_learncurve_audio_cbin_annot_notmat.toml'.
# Below, we iterate over model names
# so glob doesn't pick up static configs that are just used for testing,
# like 'invalid_option_config.toml`
TEST_CONFIGS_ROOT = TEST_DATA_ROOT.joinpath("configs")
CONFIGS_TO_RUN = []
MODELS = ("teenytweetynet", "tweetynet")
for model in MODELS:
    CONFIGS_TO_RUN.extend(sorted(TEST_CONFIGS_ROOT.glob(f"{model}*.toml")))

# the sub-directories that will get made inside `./tests/data_for_tests/generated`
TOP_LEVEL_DIRS = [
    "prep",
    "results",
]

# these sub-dirs get made in each of the TOP_LEVEL_DIRS (except for 'configs')
COMMAND_DIRS = [
    "eval",
    "learncurve",
    "predict",
    "train",
]

# these sub-dirs get made in each of the COMMAND_DIRS (except for 'configs')
DATA_DIRS = [
    "audio_cbin_annot_notmat",
    "audio_wav_annot_birdsongrec",
    "spect_mat_annot_yarden",
]


def make_subdirs_in_generated(config_paths):
    """make sub-directories inside ./tests/data_for_tests/generated

    do this after copying configs,
    before using those configs to generate results.
    We use configs to decide which dirs we need to make

    makes three directories in data_for_tests/generated:
    configs, prep, and results.
    prep has one sub-directory for every data "type".
    results does also, but in addition will have sub-directories
    within those for models.
    """
    for top_level_dir in TOP_LEVEL_DIRS:
        for command_dir in COMMAND_DIRS:
            for data_dir in DATA_DIRS:
                if not any(
                        [f'{command_dir}_{data_dir}' in str(config_path) for config_path in config_paths]
                ):
                    continue  # no need to make this dir

                if top_level_dir == "prep":
                    subdir_to_make = (
                        GENERATED_TEST_DATA / top_level_dir / command_dir / data_dir
                    )
                    subdir_to_make.mkdir(parents=True)
                else:
                    for model in MODELS:
                        subdir_to_make = (
                            GENERATED_TEST_DATA
                            / top_level_dir
                            / command_dir
                            / data_dir
                            / model
                        )
                        subdir_to_make.mkdir(parents=True)


def copy_config_files():
    """copy config files from setup to data_for_tests/configs

    the copied files are the ones that get modified when this setup script runs,
    while the originals in this directory remain unchanged.
    """
    copied_configs = []

    for toml_path in CONFIGS_TO_RUN:
        if not toml_path.exists():
            raise FileNotFoundError(f"{toml_path} not found")

        dst = GENERATED_TEST_CONFIGS_ROOT.joinpath(toml_path.name)
        print(f"\tcopying to {dst}")
        shutil.copy(src=toml_path, dst=dst)
        copied_configs.append(dst)

    return copied_configs


def run_prep(config_paths):
    """run ``vak prep`` to generate data for testing"""
    for config_path in config_paths:
        if not config_path.exists():
            raise FileNotFoundError(f"{config_path} not found")
        print(
            f"running vak prep to generate data for tests test, using config: {config_path.name}"
        )
        vak.cli.prep.prep(toml_path=config_path)


def fix_options_in_configs(config_paths, command):
    """fix values assigned to options in predict and eval configs

    Need to do this because both predict and eval configs have options
    that can only be assigned *after* running the corresponding `train` config
    """
    # split configs into train and predict or eval configs
    configs_to_fix = [config for config in config_paths if command in config.name]
    train_configs = [config for config in config_paths if "train" in config.name]

    for config_to_fix in configs_to_fix:
        # figure out which 'train' config corresponds to this 'predict' or 'eval' config
        # by using 'suffix' of config file names. `train` suffix will match `predict`/'eval' suffix
        prefix, suffix = config_to_fix.name.split(command)
        train_config_to_use = []
        for train_config in train_configs:
            train_prefix, train_suffix = train_config.name.split("train")
            if train_suffix == suffix:
                train_config_to_use.append(train_config)
        if len(train_config_to_use) != 1:
            raise ValueError(
                f"did not find just a single train config that matches with predict config:\n"
                f"{config_to_fix}"
                f"Matches were: {train_config_to_use}"
            )
        train_config_to_use = train_config_to_use[0]

        # now use the config to find the results dir and get the values for the options we need to set
        # which are checkpoint_path, spect_scaler_path, and labelmap_path
        with train_config_to_use.open("r") as fp:
            train_config_toml = toml.load(fp)
        root_results_dir = Path(train_config_toml["TRAIN"]["root_results_dir"])
        results_dir = sorted(root_results_dir.glob("results_*"))
        if len(results_dir) != 1:
            raise ValueError(
                f"did not find just a single results directory in root_results_dir from train_config:\n"
                f"{train_config_to_use}"
                f"root_results_dir was: {root_results_dir}"
                f'Matches for "results_*" were: {results_dir}'
            )
        results_dir = results_dir[0]
        # these are the only options whose values we need to change
        # and they are the same for both predict and eval
        checkpoint_path = sorted(results_dir.glob("**/checkpoints/checkpoint.pt"))[0]
        if train_config_toml['TRAIN']['normalize_spectrograms']:
            spect_scaler_path = sorted(results_dir.glob("StandardizeSpect"))[0]
        else:
            spect_scaler_path = None
        labelmap_path = sorted(results_dir.glob("labelmap.json"))[0]

        # now add these values to corresponding options in predict / eval config
        with config_to_fix.open("r") as fp:
            config_toml = toml.load(fp)
        if command == 'train_continue':
            section = 'TRAIN'
        else:
            section = command.upper()
        config_toml[section]["checkpoint_path"] = str(checkpoint_path)
        if spect_scaler_path:
            config_toml[section]["spect_scaler_path"] = str(spect_scaler_path)
        else:
            if 'spect_scaler_path' in config_toml[section]:
                # remove any existing 'spect_scaler_path' option
                del config_toml[section]["spect_scaler_path"]
        config_toml[section]["labelmap_path"] = str(labelmap_path)
        with config_to_fix.open("w") as fp:
            toml.dump(config_toml, fp)


# need to run 'train' config before we run 'predict'
# so we can add checkpoints, etc., from training to predict
COMMANDS = (
    "train",
    "learncurve",
    "eval",
    "predict",
    "train_continue",
)


def main():
    print(
        "copying config files run to generate test data from ./tests/data_for_tests/configs to "
        "./tests/data_for_tests/generated/configs"
    )
    GENERATED_TEST_CONFIGS_ROOT.mkdir(parents=True)
    config_paths = copy_config_files()

    print(f"will generate test data from these config files: {config_paths}")

    print(
        "making sub-directories in ./tests/data_for_tests/generated/ where files generated by `vak` will go"
    )
    make_subdirs_in_generated(config_paths)

    # ---- only need to run prep once, since prep'd data is the same regardless of model ----
    prep_config_paths = [
        config_path
        for config_path in config_paths
        if config_path.name.startswith(MODELS[0])
    ]
    run_prep(config_paths=prep_config_paths)
    # now add the prep csv from those configs to the corresponding config
    # from all the other models
    for model in MODELS[1:]:
        model_config_paths = [
            config_path
            for config_path in config_paths
            if config_path.name.startswith(model)
        ]
        for model_config_path in model_config_paths:
            # we want the same prep config for MODEL[0] which will have the
            # exact same name, but with a different model name as the "prefix"
            stem_minus_model = model_config_path.stem.replace(model, "")
            prep_config_path = [
                prep_config_path
                for prep_config_path in prep_config_paths
                if prep_config_path.stem.endswith(stem_minus_model)
            ]
            assert len(prep_config_path) == 1
            prep_config_path = prep_config_path[0]
            with prep_config_path.open("r") as fp:
                prep_config_toml = toml.load(fp)
            with model_config_path.open("r") as fp:
                model_config_toml = toml.load(fp)
            # find the section that `vak prep` added the `csv_path` to,
            # and set `csv_path` for model config to the same value in
            # the same section for this model config
            for section_name, options_dict in prep_config_toml.items():
                if "csv_path" in options_dict:
                    model_config_toml[section_name]["csv_path"] = options_dict[
                        "csv_path"
                    ]
            with model_config_path.open("w") as fp:
                toml.dump(model_config_toml, fp)

    for model in MODELS:
        for command in COMMANDS:
            if command == "prep":
                continue  # already ran 'prep'
            print(f"running configs for command: {command}")
            command_config_paths = [
                config_path
                for config_path in config_paths
                if config_path.name.startswith(model) and command in config_path.name
            ]
            if command == "train":
                # need to remove 'train_continue' configs
                command_config_paths = [
                    config_path for config_path in command_config_paths 
                    if 'continue' not in config_path.name
                ]
            elif command == "train_continue":
                # only keep 'train_continue' configs
                command_config_paths = [
                    config_path for config_path in command_config_paths 
                    if 'continue' in config_path.name
                ]
            print(f"using the following configs:\n{command_config_paths}")
            if command in ("predict", "eval", "train_continue"):
                # fix values for required options in predict / eval configs
                # using results from running the corresponding train configs.
                # this only works if we ran the train configs already,
                # which we should have because of ordering of COMMANDS constant above
                copied_config_paths_this_model = [
                    config_path
                    for config_path in config_paths
                    if config_path.name.startswith(model)
                ]
                fix_options_in_configs(copied_config_paths_this_model, command)

            if command == 'train' or command == 'learncurve':
                # we run `train` to get results needed for `eval', 'predict' and continuing 'train';
                # we run `learncurve` so there's a `previous_run_path` to test;
                # skip all other commands
                for config_path in command_config_paths:
                    vak.cli.cli.cli(command, config_path)


if __name__ == "__main__":
    main()



import re

import numpy as np
import pandas as pd

from . import train_dur_csv_paths as _train_dur_csv_paths
from ..eval import eval
from ..train import train
from ...io import dataframe
from ... import csv, labels
from ...converters import expanded_user_path
from ...logging import log_or_print
from ...paths import generate_results_dir_name_as_path


def learning_curve(
    model_config_map,
    train_set_durs,
    num_replicates,
    csv_path,
    labelset,
    window_size,
    batch_size,
    num_epochs,
    num_workers,
    root_results_dir=None,
    results_path=None,
    previous_run_path=None,
    spect_key="s",
    timebins_key="t",
    normalize_spectrograms=True,
    shuffle=True,
    val_step=None,
    ckpt_step=None,
    patience=None,
    device=None,
    logger=None,
):
    """generate learning curve, by training models on training sets across a
    range of sizes and then measure accuracy of those models on a test set.

    Parameters
    ----------
    model_config_map : dict
        where each key-value pair is model name : dict of config parameters
    train_set_durs : list
        of int, durations in seconds of subsets taken from training data
        to create a learning curve, e.g. [5, 10, 15, 20].
    num_replicates : int
        number of times to replicate training for each training set duration
        to better estimate metrics for a training set of that size.
        Each replicate uses a different randomly drawn subset of the training
        data (but of the same duration).
    csv_path : str
        path to where dataset was saved as a csv.
    labelset : set
        of str or int, the set of labels that correspond to annotated segments
        that a network should learn to segment and classify. Note that if there
        are segments that are not annotated, e.g. silent gaps between songbird
        syllables, then `vak` will assign a dummy label to those segments
        -- you don't have to give them a label here.
    window_size : int
        size of windows taken from spectrograms, in number of time bins,
        shonw to neural networks
    batch_size : int
        number of samples per batch presented to models during training.
    num_epochs : int
        number of training epochs. One epoch = one iteration through the entire
        training set.
    num_workers : int
        Number of processes to use for parallel loading of data.
        Argument to torch.DataLoader.
    root_results_dir : str, pathlib.Path
        Root directory in which a new directory will be created where results will be saved.
    results_path : str, pathlib.Path
        Directory where results will be saved. If specified, this parameter overrides root_results_dir.
    previous_run_path : str, Path
        path to directory containing dataset .csv files
        that represent subsets of training set, created by
        a previous run of ``vak.core.learncurve.learning_curve``.
        Typically directory will have a name like ``results_{timestamp}``
        and the actual .csv splits will be in sub-directories with names
        corresponding to the training set duration
    spect_key : str
        key for accessing spectrogram in files. Default is 's'.
    timebins_key : str
        key for accessing vector of time bins in files. Default is 't'.
    device : str
        Device on which to work with model + data.
        Default is None. If None, then a device will be selected with vak.device.get_default.
        That function defaults to 'cuda' if torch.cuda.is_available is True.
    shuffle: bool
        if True, shuffle training data before each epoch. Default is True.
    normalize_spectrograms : bool
        if True, use spect.utils.data.SpectScaler to normalize the spectrograms.
        Normalization is done by subtracting off the mean for each frequency bin
        of the training set and then dividing by the std for that frequency bin.
        This same normalization is then applied to validation + test data.
    val_step : int
        Step on which to estimate accuracy using validation set.
        If val_step is n, then validation is carried out every time
        the global step / n is a whole number, i.e., when val_step modulo the global step is 0.
        Default is None, in which case no validation is done.
    ckpt_step : int
        Step on which to save to checkpoint file.
        If ckpt_step is n, then a checkpoint is saved every time
        the global step / n is a whole number, i.e., when ckpt_step modulo the global step is 0.
        Default is None, in which case checkpoint is only saved at the last epoch.
    patience : int
        number of validation steps to wait without performance on the
        validation set improving before stopping the training.
        Default is None, in which case training only stops after the specified number of epochs.

    Other Parameters
    ----------------
    logger : logging.Logger
        instance created by ``vak.logging.get_logger``. Default is None.

    Returns
    -------
    None

    Trains models, saves results in new directory within root_results_dir
    """
    # ---------------- pre-conditions ----------------------------------------------------------------------------------
    csv_path = expanded_user_path(csv_path)
    if not csv_path.exists():
        raise FileNotFoundError(f"csv_path not found: {csv_path}")

    log_or_print(f"Using dataset from .csv: {csv_path}", logger=logger, level="info")
    dataset_df = pd.read_csv(csv_path)

    if previous_run_path:
        previous_run_path = expanded_user_path(previous_run_path)
        if not previous_run_path.is_dir():
            raise NotADirectoryError(
                f"previous_run_path not recognized as a directory:\n{previous_run_path}"
            )

    if val_step and not dataset_df["split"].str.contains("val").any():
        raise ValueError(
            f"val_step set to {val_step} but dataset does not contain a validation set; "
            f"please run `vak prep` with a config.toml file that specifies a duration for the validation set."
        )

    # ---- set up directory to save output -----------------------------------------------------------------------------
    if results_path:
        results_path = expanded_user_path(results_path)
        if not results_path.is_dir():
            raise NotADirectoryError(
                f"results_path not recognized as a directory: {results_path}"
            )
    else:
        results_path = generate_results_dir_name_as_path(root_results_dir)
        results_path.mkdir()

    log_or_print(f"Saving results to: {results_path}", logger=logger, level="info")

    timebin_dur = dataframe.validate_and_get_timebin_dur(dataset_df)
    log_or_print(
        f"Size of each timebin in spectrogram, in seconds: {timebin_dur}",
        logger=logger,
        level="info",
    )

    # ---- get training set subsets ------------------------------------------------------------------------------------
    has_unlabeled = csv.has_unlabeled(csv_path, labelset, timebins_key)
    if has_unlabeled:
        map_unlabeled = True
    else:
        map_unlabeled = False
    labelmap = labels.to_map(labelset, map_unlabeled=map_unlabeled)

    if previous_run_path:
        log_or_print(
            f"Loading previous training subsets from:\n{previous_run_path}",
            logger=logger,
            level="info",
        )
        train_dur_csv_paths = _train_dur_csv_paths.from_dir(
            previous_run_path,
            train_set_durs,
            timebin_dur,
            num_replicates,
            results_path,
            window_size,
            spect_key,
            timebins_key,
            labelmap,
        )
    else:
        log_or_print(
            f"Creating data sets of specified durations: {train_set_durs}",
            logger=logger,
            level="info",
        )
        # do all subsetting before training, so that we fail early if subsetting is going to fail
        train_dur_csv_paths = _train_dur_csv_paths.from_df(
            dataset_df,
            csv_path,
            train_set_durs,
            timebin_dur,
            num_replicates,
            results_path,
            labelset,
            window_size,
            spect_key,
            timebins_key,
            labelmap,
            logger,
        )

    # ---- main loop that creates "learning curve" ---------------------------------------------------------------------
    log_or_print(f"Starting training for learning curve.", logger=logger, level="info")
    for train_dur, csv_paths in train_dur_csv_paths.items():
        log_or_print(
            f"Training replicates for training set of size: {train_dur}s",
            logger=logger,
            level="info",
        )

        for replicate_num, this_train_dur_this_replicate_csv_path in enumerate(
            csv_paths
        ):
            replicate_num += 1  # so log statements below match replicate nums returned by train_dur_csv_paths
            log_or_print(
                f"Training replicate {replicate_num} "
                f"using dataset from .csv file: {this_train_dur_this_replicate_csv_path}",
                logger=logger,
                level="info",
            )
            this_train_dur_this_replicate_results_path = (
                this_train_dur_this_replicate_csv_path.parent
            )
            log_or_print(
                f"Saving results to: {this_train_dur_this_replicate_results_path}",
                logger=logger,
                level="info",
            )

            window_dataset_kwargs = {}
            for window_dataset_kwarg in [
                "spect_id_vector",
                "spect_inds_vector",
                "x_inds",
            ]:
                window_dataset_kwargs[window_dataset_kwarg] = np.load(
                    this_train_dur_this_replicate_results_path.joinpath(
                        f"{window_dataset_kwarg}.npy"
                    )
                )

            train(
                model_config_map,
                this_train_dur_this_replicate_csv_path,
                window_size,
                batch_size,
                num_epochs,
                num_workers,
                labelset=labelset,
                results_path=this_train_dur_this_replicate_results_path,
                spect_key=spect_key,
                timebins_key=timebins_key,
                normalize_spectrograms=normalize_spectrograms,
                shuffle=shuffle,
                val_step=val_step,
                ckpt_step=ckpt_step,
                patience=patience,
                device=device,
                logger=logger,
                **window_dataset_kwargs,
            )

            log_or_print(
                f"Evaluating models from replicate {replicate_num} "
                f"using dataset from .csv file: {this_train_dur_this_replicate_results_path}",
                logger=logger,
                level="info",
            )
            for model_name in model_config_map.keys():
                log_or_print(
                    f"Evaluating model: {model_name}", logger=logger, level="info"
                )
                results_model_root = (
                    this_train_dur_this_replicate_results_path.joinpath(model_name)
                )
                ckpt_root = results_model_root.joinpath("checkpoints")
                ckpt_paths = sorted(ckpt_root.glob("*.pt"))
                if any(["max-val-acc" in str(ckpt_path) for ckpt_path in ckpt_paths]):
                    ckpt_paths = [
                        ckpt_path
                        for ckpt_path in ckpt_paths
                        if "max-val-acc" in str(ckpt_path)
                    ]
                    if len(ckpt_paths) != 1:
                        raise ValueError(
                            f"did not find a single max-val-acc checkpoint path, instead found:\n{ckpt_paths}"
                        )
                    ckpt_path = ckpt_paths[0]
                else:
                    if len(ckpt_paths) != 1:
                        raise ValueError(
                            f"did not find a single checkpoint path, instead found:\n{ckpt_paths}"
                        )
                    ckpt_path = ckpt_paths[0]
                log_or_print(
                    f"Using checkpoint: {ckpt_path}", logger=logger, level="info"
                )
                labelmap_path = this_train_dur_this_replicate_results_path.joinpath(
                    "labelmap.json"
                )
                log_or_print(
                    f"Using labelmap: {labelmap_path}", logger=logger, level="info"
                )
                if normalize_spectrograms:
                    spect_scaler_path = (
                        this_train_dur_this_replicate_results_path.joinpath(
                            "StandardizeSpect"
                        )
                    )
                    log_or_print(
                        f"Using spect scaler to normalize: {spect_scaler_path}",
                        logger=logger,
                        level="info",
                    )
                else:
                    spect_scaler_path = None

                eval(
                    this_train_dur_this_replicate_csv_path,
                    model_config_map,
                    checkpoint_path=ckpt_path,
                    labelmap_path=labelmap_path,
                    output_dir=this_train_dur_this_replicate_results_path,
                    window_size=window_size,
                    num_workers=num_workers,
                    split="test",
                    spect_scaler_path=spect_scaler_path,
                    spect_key=spect_key,
                    timebins_key=timebins_key,
                    device=device,
                    logger=logger,
                )

    # ---- make a csv for analysis -------------------------------------------------------------------------------------
    reg_exp_num = re.compile(
        r"[-+]?\d*\.\d+|\d+"
    )  # to extract train set dur and replicate num from paths

    eval_csv_paths = sorted(results_path.glob("**/eval*.csv"))
    eval_df_0 = pd.read_csv(eval_csv_paths[0])  # use to just get columns
    eval_columns = eval_df_0.columns.tolist()  # will use below to re-order
    eval_dfs = []
    for eval_csv_path in eval_csv_paths:
        train_set_dur = reg_exp_num.findall(eval_csv_path.parents[1].name)
        if len(train_set_dur) != 1:
            raise ValueError(
                f"unable to determine training set duration from .csv path: {train_set_dur}"
            )
        else:
            train_set_dur = float(train_set_dur[0])
        replicate_num = reg_exp_num.findall(eval_csv_path.parents[0].name)
        if len(replicate_num) != 1:
            raise ValueError(
                f"unable to determine replicate number from .csv path: {train_set_dur}"
            )
        else:
            replicate_num = int(replicate_num[0])
        eval_df = pd.read_csv(eval_csv_path)
        eval_df["train_set_dur"] = train_set_dur
        eval_df["replicate_num"] = replicate_num
        eval_dfs.append(eval_df)
    all_eval_df = pd.concat(eval_dfs)
    all_eval_columns = ["train_set_dur", "replicate_num", *eval_columns]
    all_eval_df = all_eval_df[all_eval_columns]
    all_eval_df.sort_values(by=["train_set_dur", "replicate_num"])
    learncurve_csv_path = results_path.joinpath("learning_curve.csv")
    all_eval_df.to_csv(
        learncurve_csv_path, index=False
    )  # index=False to avoid adding "Unnamed: 0" column



import pytest


@pytest.fixture
def default_model():
    """default model used whenever a model is needed to run a test.
    Should work regardless of where the test is run, i.e. both on
    CI platform and locally.

    currently ``teenytweetynet``
    """
    return "teenytweetynet"



"""assertions common to cli functions"""


def toml_config_file_copied_to_results_path(results_path, toml_path):
    assert results_path.joinpath(toml_path.name).exists()
    return True


def log_file_created(command, output_path):
    log_file = sorted(output_path.glob(f"{command}*.log"))
    assert len(log_file) == 1

    return True



"""tests for vak.config.learncurve module"""
import vak.config.learncurve
import vak.split


def test_learncurve_attrs_class(all_generated_learncurve_configs_toml):
    """test that instantiating LearncurveConfig class works as expected"""
    for config_toml in all_generated_learncurve_configs_toml:
        learncurve_section = config_toml["LEARNCURVE"]
        config = vak.config.learncurve.LearncurveConfig(**learncurve_section)
        assert isinstance(config, vak.config.learncurve.LearncurveConfig)



"""functions dealing with ``tensorboard``"""
from pathlib import Path

import pandas as pd
from torch.utils.tensorboard import SummaryWriter
from tensorboard.backend.event_processing.event_accumulator import EventAccumulator


def get_summary_writer(log_dir, filename_suffix):
    """get an instance of ``tensorboard.SummaryWriter``,
    to use with a vak.Model during training

    Parameters
    ----------
    log_dir : str
        directory where event file will be written
    filename_suffix : str
        suffix added to events file name

    Returns
    -------
    summary_writer : torch.utils.tensorboard.SummaryWriter

    Examples
    --------
    >>> summary_writer = vak.summary_writer.get_summary_writer(log_dir='./experiments')
    >>> tweety_net_model.summary_writer = summary_writer  # set attribute equal to instance we just made
    >>> tweety_net_model.train()  # now events during training will be logged with that summary writer
    """
    return SummaryWriter(log_dir=log_dir, filename_suffix=filename_suffix)


DEFAULT_SIZE_GUIDANCE = {
    "compressedHistograms": 1,
    "images": 1,
    "scalars": 0,  # 0 means load all
    "histograms": 1,
}


def events2df(events_path, size_guidance=None, drop_wall_time=True):
    """convert ``tensorboard`` "events" log file to pandas DataFrame

    events files are created by SummaryWriter from PyTorch or Tensorflow.

    Parameters
    ----------
    events_path : str, Path
        path to either a log directory or a specific events file
        saved by a SummaryWriter in a log directory.
        By default, ``vak`` saves logs in a directory with the model name
        inside a ``results`` directory generated at the start of training.
    size_guidance: dict
        Argument passed to the ``EventAccumlator`` class from
        ``tensorboard`` that is used to load the events file.
        Information on how much data the EventAccumulator should
        store in memory. Dict that maps a `tagType` string
        to an integer representing the number of items to keep per tag
        for items of that `tagType`. If the size is 0,
        all events are stored. Default is None, in which case
        ``vak.tensorboard.DEFAULT_SIZE_GUIDANCE`` is used.
        For more information see
        https://github.com/tensorflow/tensorboard/blob/master/tensorboard/backend/event_processing/event_accumulator.py
    drop_wall_time : bool
        if True, drop wall times logged in events file. Default is True.

    Returns
    -------
    df : pandas.Dataframe
        with index 'step' and all Scalars from the events file

    Examples
    --------
    >>> events_path = 'tweetynet/results_210322_103904/train_dur_6s/replicate_2/TweetyNet/'
    >>> events_df = vak.tensorboard.events2df(events_path)
    >>> events_df
              loss/train  avg_acc/val  avg_levenshtein/val  avg_segment_error_rate/val  avg_loss/val
    step
    0       2.479142          NaN                  NaN                         NaN           NaN
    1       2.458833          NaN                  NaN                         NaN           NaN
    2       2.441571          NaN                  NaN                         NaN           NaN
    3       2.402737          NaN                  NaN                         NaN           NaN
    4       2.404369          NaN                  NaN                         NaN           NaN
    ...          ...          ...                  ...                         ...           ...
    996     0.171681          NaN                  NaN                         NaN           NaN
    997     0.100202          NaN                  NaN                         NaN           NaN
    998     0.073055          NaN                  NaN                         NaN           NaN
    999     0.031479          NaN                  NaN                         NaN           NaN
    1000         NaN     0.902475                 42.0                    0.880533      0.310385

    [1001 rows x 5 columns]
    """
    if isinstance(events_path, Path):
        events_path = str(events_path)

    if size_guidance is None:
        size_guidance = DEFAULT_SIZE_GUIDANCE

    ea = EventAccumulator(path=events_path, size_guidance=size_guidance)
    ea.Reload()  # load all data written so far

    scalar_tags = ea.Tags()["scalars"]  # list of tags for values written to scalar
    # make a dataframe for each tag, which we will then concatenate using 'step' as the index
    # so that pandas will fill in with NaNs for any scalars that were not measured on every step
    dfs = {}
    for scalar_tag in scalar_tags:
        dfs[scalar_tag] = pd.DataFrame(
            ea.Scalars(scalar_tag), columns=["wall_time", "step", scalar_tag]
        )
        dfs[scalar_tag] = dfs[scalar_tag].set_index("step")
        if drop_wall_time:
            dfs[scalar_tag].drop("wall_time", axis=1, inplace=True)
    return pd.concat([v for k, v in dfs.items()], axis=1)



"""module that handles file input-output:
- audio files
- spectrograms made from audio files of vocalizations
- .csv files that represent a dataset of vocalizations that combines all those files together"""
from . import audio, dataframe, spect



# adopted from attrs library
# https://github.com/python-attrs/attrs/blob/master/tests/utils.py
from attr import Attribute
from attr._make import NOTHING


def simple_attr(
    name,
    default=NOTHING,
    validator=None,
    repr=True,
    eq=True,
    hash=None,
    init=True,
    converter=None,
    kw_only=False,
    inherited=False,
):
    """
    Return an attribute with a name and no other bells and whistles.
    """
    return Attribute(
        name=name,
        default=default,
        validator=validator,
        repr=repr,
        cmp=None,
        eq=eq,
        hash=hash,
        init=init,
        converter=converter,
        kw_only=kw_only,
        inherited=inherited,
    )



"""tests for vak.split.algorithms.bruteforce module"""
from math import isclose

from vak.split.algorithms import brute_force

# since the algorithm is random, we test multiple times
# yes, this opens up the possibility of non-deterministic failures
NUM_SAMPLES = 10


def is_expected_output(
    train_dur,
    val_dur,
    test_dur,
    labelset,
    durs,
    labels,
    train_inds,
    val_inds,
    test_inds,
):
    """asserts that output from ``bruteforce`` is expected output

    Checks that each split returned by ``bruteforce`` is the
    expected duration, and contains all the labels in ``labelset``.

    Parameters
    ----------
    train_dur
    val_dur
    test_dur
    labelset
    durs
    labels
    train_inds
    val_inds
    test_inds

    Returns
    -------
    expected : bool
        True if output matched expected

    Notes
    -----
    uses specified target durations for splits
    (``train_dur``, ``val_dur``, and ``test_dur``)
    and list of indices that correspond to splits
    found by ``split.algorithms.bruteforce``
    (``train_inds``, ``val_inds``, and ``test_inds``)
    to check that durations of returned splits are:
    - equal to or greater than specified target duration,
      when target duration > 0
    - equal to the total duration of the dataset minus
      the size of the other split,
      when the target duration is set to -1.
      E.g., if the target ``test_dur`` is set to -1,
      the duration of the returned split should be
      approximately the total duration of the dataset
      minus the duration of the returned ``train_inds``.
    When a duration is specified as None, then this function
    asserts that ``bruteforce`` returned None.

    For any split with a target duration (not None),
    this function also checks that the set of labels
    in the split equals the specified ``labelset``.
    """
    for split, dur_in, inds in zip(
        ("train", "val", "test"),
        (train_dur, val_dur, test_dur),
        (train_inds, val_inds, test_inds),
    ):
        if dur_in is not None:
            dur_out = sum([durs[ind] for ind in inds])
            if dur_in > 0:
                assert dur_out >= dur_in
            elif dur_in == -1:
                if split == "train":
                    assert isclose(
                        dur_out, sum(durs) - sum([durs[ind] for ind in test_inds])
                    )
                elif split == "test":
                    assert isclose(
                        dur_out, sum(durs) - sum([durs[ind] for ind in train_inds])
                    )

            all_lbls_this_set = [lbl for ind in inds for lbl in labels[ind]]
            assert labelset == set(all_lbls_this_set)
        else:
            assert inds is None

    assert set(train_inds).isdisjoint(set(test_inds))
    if val_dur is not None:
        assert set(train_inds).isdisjoint(set(val_inds))
        assert set(test_inds).isdisjoint(set(val_inds))

    return True


def test_bruteforce_train_test_val_mock():
    train_dur = 2
    test_dur = 2
    val_dur = 1
    durs = (1, 1, 1, 1, 1)
    labelset = set(list("abcde"))
    labels = [list("abcde") for _ in range(5)]

    for _ in range(NUM_SAMPLES):
        train_inds, val_inds, test_inds = brute_force(
            durs, labels, labelset, train_dur, val_dur, test_dur
        )

        assert is_expected_output(
            train_dur,
            val_dur,
            test_dur,
            labelset,
            durs,
            labels,
            train_inds,
            val_inds,
            test_inds,
        )


def test_bruteforce_train_test_val_cbin(
    audio_cbin_annot_notmat_durs_labels, labelset_notmat
):
    labelset_notmat = set(labelset_notmat)
    durs, labels = audio_cbin_annot_notmat_durs_labels
    train_dur = 35
    val_dur = 20
    test_dur = 35

    for _ in range(NUM_SAMPLES):
        train_inds, val_inds, test_inds = brute_force(
            durs, labels, labelset_notmat, train_dur, val_dur, test_dur
        )

        assert is_expected_output(
            train_dur,
            val_dur,
            test_dur,
            labelset_notmat,
            durs,
            labels,
            train_inds,
            val_inds,
            test_inds,
        )


def test_bruteforce_train_test_val_mat(
    spect_mat_annot_yarden_durs_labels, labelset_yarden
):
    labelset_yarden = set(labelset_yarden)
    durs, labels = spect_mat_annot_yarden_durs_labels
    train_dur = 200
    val_dur = 100
    test_dur = 200

    for _ in range(NUM_SAMPLES):
        train_inds, val_inds, test_inds = brute_force(
            durs, labels, labelset_yarden, train_dur, val_dur, test_dur
        )

        assert is_expected_output(
            train_dur,
            val_dur,
            test_dur,
            labelset_yarden,
            durs,
            labels,
            train_inds,
            val_inds,
            test_inds,
        )


def test_bruteforce_train_neg_one_test_mock():
    train_dur = 2
    val_dur = None
    test_dur = -1
    durs = (1, 1, 1, 1, 1)
    labelset = set(list("abcde"))
    labels = [list("abcde") for _ in range(5)]

    for _ in range(NUM_SAMPLES):
        train_inds, val_inds, test_inds = brute_force(
            durs, labels, labelset, train_dur, val_dur, test_dur
        )

        assert is_expected_output(
            train_dur,
            val_dur,
            test_dur,
            labelset,
            durs,
            labels,
            train_inds,
            val_inds,
            test_inds,
        )


def test_bruteforce_train_neg_one_test_cbin(
    audio_cbin_annot_notmat_durs_labels, labelset_notmat
):
    labelset_notmat = set(labelset_notmat)
    durs, labels = audio_cbin_annot_notmat_durs_labels
    train_dur = 35
    val_dur = None
    test_dur = -1

    for _ in range(NUM_SAMPLES):
        train_inds, val_inds, test_inds = brute_force(
            durs, labels, labelset_notmat, train_dur, val_dur, test_dur
        )

        assert is_expected_output(
            train_dur,
            val_dur,
            test_dur,
            labelset_notmat,
            durs,
            labels,
            train_inds,
            val_inds,
            test_inds,
        )


def test_bruteforce_train_mat(spect_mat_annot_yarden_durs_labels, labelset_yarden):
    labelset_yarden = set(labelset_yarden)
    durs, labels = spect_mat_annot_yarden_durs_labels
    train_dur = 300
    val_dur = None
    test_dur = -1

    for _ in range(NUM_SAMPLES):
        train_inds, val_inds, test_inds = brute_force(
            durs, labels, labelset_yarden, train_dur, val_dur, test_dur
        )

        assert is_expected_output(
            train_dur,
            val_dur,
            test_dur,
            labelset_yarden,
            durs,
            labels,
            train_inds,
            val_inds,
            test_inds,
        )


def test_bruteforce_test_mock():
    train_dur = -1
    test_dur = 2
    val_dur = None
    durs = (1, 1, 1, 1, 1)
    labelset = set(list("abcde"))
    labels = [list("abcde") for _ in range(5)]

    for _ in range(NUM_SAMPLES):
        train_inds, val_inds, test_inds = brute_force(
            durs, labels, labelset, train_dur, val_dur, test_dur
        )

        assert is_expected_output(
            train_dur,
            val_dur,
            test_dur,
            labelset,
            durs,
            labels,
            train_inds,
            val_inds,
            test_inds,
        )


def test_bruteforce_test_cbin(audio_cbin_annot_notmat_durs_labels, labelset_notmat):
    labelset_notmat = set(labelset_notmat)
    durs, labels = audio_cbin_annot_notmat_durs_labels
    train_dur = -1
    val_dur = None
    test_dur = 25

    for _ in range(NUM_SAMPLES):
        train_inds, val_inds, test_inds = brute_force(
            durs, labels, labelset_notmat, train_dur, val_dur, test_dur
        )

        assert is_expected_output(
            train_dur,
            val_dur,
            test_dur,
            labelset_notmat,
            durs,
            labels,
            train_inds,
            val_inds,
            test_inds,
        )


def test_bruteforce_test_mat(spect_mat_annot_yarden_durs_labels, labelset_yarden):
    labelset_yarden = set(labelset_yarden)
    durs, labels = spect_mat_annot_yarden_durs_labels
    train_dur = -1
    val_dur = None
    test_dur = 200

    for _ in range(NUM_SAMPLES):
        train_inds, val_inds, test_inds = brute_force(
            durs, labels, labelset_yarden, train_dur, val_dur, test_dur
        )

        assert is_expected_output(
            train_dur,
            val_dur,
            test_dur,
            labelset_yarden,
            durs,
            labels,
            train_inds,
            val_inds,
            test_inds,
        )
